#
# Copyright (c) 2021, self.experiment Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
* self.experiment_API_TOKEN
* self.experiment_PROJECT
"""
import os
import sys
from datetime import datetime

from neptune import Session, envs
from common_client_code import ClientFeatures


class OldClientNonglobalFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        self.session = Session()
        self.project = self.session.get_project(os.getenv(envs.PROJECT_ENV_NAME))
        self.experiment = self.project.create_experiment(
            name='const project name',
            params=self.params,
            tags=['initial tag 1', 'initial tag 2'],
        )

    def modify_tags(self):
        self.experiment.append_tags('tag1')
        self.experiment.append_tag(['tag2_to_remove', 'tag3'])
        # self.experiment.remove_tag('tag2_to_remove')  # TODO: NPT-9222
        # self.experiment.remove_tag('tag4_remove_non_existing')  # TODO: NPT-9222

        exp = self.experiment
        assert set(exp.get_tags()) == {'initial tag 1', 'initial tag 2', 'tag1', 'tag2_to_remove', 'tag3'}

    def modify_properties(self):
        self.experiment.set_property('prop', 'some text')
        self.experiment.set_property('prop_number', 42)
        self.experiment.set_property('nested/prop', 42)
        self.experiment.set_property('prop_to_del', 42)
        self.experiment.set_property('prop_list', [1, 2, 3])
        with open(self.text_file_path, mode='r') as f:
            self.experiment.set_property('prop_IO', f)
        self.experiment.set_property('prop_datetime', datetime.now())
        self.experiment.remove_property('prop_to_del')

        properties = self.experiment.get_properties()
        assert properties['prop'] == 'some text'
        assert properties['prop_number'] == '42'
        assert properties['nested/prop'] == '42'
        assert 'prop_to_del' not in properties
        assert properties['prop_IO'] == "<_io.TextIOWrapper name='alpha_integration_dev/data/text.txt'" \
                                        " mode='r' encoding='UTF-8'>"
        print(f'Properties: {properties}')

    def log_std(self):
        print('stdout text1')
        print('stdout text2')
        print('stderr text1', file=sys.stderr)
        print('stderr text2', file=sys.stderr)

    def log_series(self):
        # floats
        self.experiment.log_metric('m1', 1)
        self.experiment.log_metric('m1', 2)
        self.experiment.log_metric('m1', 3)
        self.experiment.log_metric('m1', 2)
        self.experiment.log_metric('nested/m1', 1)

        # texts
        self.experiment.log_text('m2', 'a')
        self.experiment.log_text('m2', 'b')
        self.experiment.log_text('m2', 'c')

        # images
        # `image_name` and `description` will be lost
        self.experiment.log_image('g_img', self.img_path, image_name='name', description='desc')
        self.experiment.log_image('g_img', self.img_path)

        # see what we've logged
        logs = self.experiment.get_logs()
        print(f'Logs: {logs}')

    def handle_files_and_images(self):
        # image
        # `image_name` and `description` will be lost (`send_image` the same as `log_image`)
        self.experiment.send_image('image', self.img_path, name='name', description='desc')

        # artifact
        # (`log_artifact` the same as `log_artifact`)
        self.experiment.send_artifact(self.text_file_path)
        self.experiment.log_artifact(self.text_file_path, destination='dir/text file artifact')
        with open(self.text_file_path, mode='r') as f:
            self.experiment.send_artifact(f, destination='file stream.txt')
        self.experiment.log_artifact(self.img_path, destination='dir to delete/art1')
        self.experiment.log_artifact(self.img_path, destination='dir to delete/art2')
        # self.experiment.delete_artifacts('dir to delete')  # doesn't work for alpha NPT-9250
        self.experiment.delete_artifacts('dir to delete/art1')

    def finalize(self):
        pass


if __name__ == '__main__':
    OldClientNonglobalFeatures().run()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

ARTIFACT_ATTRIBUTE_SPACE = 'artifacts/'

LOG_ATTRIBUTE_SPACE = 'logs/'

MONITORING_ATTRIBUTE_SPACE = 'monitoring/'
MONITORING_STDERR_ATTRIBUTE_PATH = f'{MONITORING_ATTRIBUTE_SPACE}stderr'
MONITORING_STDOUT_ATTRIBUTE_PATH = f'{MONITORING_ATTRIBUTE_SPACE}stdout'
MONITORING_TRACEBACK_ATTRIBUTE_PATH = f'{MONITORING_ATTRIBUTE_SPACE}traceback'

PARAMETERS_ATTRIBUTE_SPACE = 'parameters/'

PROPERTIES_ATTRIBUTE_SPACE = 'properties/'

SOURCE_CODE_ATTRIBUTE_SPACE = 'source_code/'
SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH = f'{SOURCE_CODE_ATTRIBUTE_SPACE}entrypoint'
SOURCE_CODE_FILES_ATTRIBUTE_PATH = f'{SOURCE_CODE_ATTRIBUTE_SPACE}files'

SYSTEM_ATTRIBUTE_SPACE = 'sys/'
SYSTEM_DESCRIPTION_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}description'
SYSTEM_HOSTNAME_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}hostname'
SYSTEM_NAME_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}name'
SYSTEM_STATE_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}state'
SYSTEM_TAGS_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}tags'
SYSTEM_FAILED_ATTRIBUTE_PATH = f'{SYSTEM_ATTRIBUTE_SPACE}failed'

SIGNAL_TYPE_STOP = "neptune/stop"
SIGNAL_TYPE_ABORT = "neptune/abort"



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os
import threading
from datetime import datetime
from enum import Enum
from pathlib import Path
from platform import node as get_hostname
from typing import List, Optional, Union

import click

from neptune.new.attributes import constants as attr_consts
from neptune.new.constants import (
    ASYNC_DIRECTORY,
    NEPTUNE_RUNS_DIRECTORY,
    OFFLINE_DIRECTORY,
)
from neptune.new.envs import (CUSTOM_RUN_ID_ENV_NAME, NEPTUNE_NOTEBOOK_ID, NEPTUNE_NOTEBOOK_PATH,
                              MONITORING_NAMESPACE)
from neptune.new.exceptions import (NeedExistingRunForReadOnlyMode, NeptuneRunResumeAndCustomIdCollision,
                                    NeptunePossibleLegacyUsageException)
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.backends.project_name_lookup import project_name_lookup
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.new.internal.backends.offline_neptune_backend import OfflineNeptuneBackend
from neptune.new.internal.backgroud_job_list import BackgroundJobList
from neptune.new.internal.containers.disk_queue import DiskQueue
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.hardware.hardware_metric_reporting_job import HardwareMetricReportingJob
from neptune.new.internal.notebooks.notebooks import create_checkpoint
from neptune.new.internal.operation import Operation
from neptune.new.internal.operation_processors.async_operation_processor import AsyncOperationProcessor
from neptune.new.internal.operation_processors.offline_operation_processor import OfflineOperationProcessor
from neptune.new.internal.operation_processors.read_only_operation_processor import ReadOnlyOperationProcessor
from neptune.new.internal.operation_processors.sync_operation_processor import SyncOperationProcessor
from neptune.new.internal.streams.std_capture_background_job import (
    StderrCaptureBackgroundJob,
    StdoutCaptureBackgroundJob,
)
from neptune.new.internal.utils import verify_collection_type, verify_type
from neptune.new.internal.utils.git import discover_git_repo_location, get_git_info
from neptune.new.internal.utils.ping_background_job import PingBackgroundJob
from neptune.new.internal.utils.runningmode import in_interactive, in_notebook
from neptune.new.internal.utils.source_code import upload_source_code
from neptune.new.internal.utils.traceback_job import TracebackJob
from neptune.new.internal.utils.uncaught_exception_handler import instance as uncaught_exception_handler
from neptune.new.internal.websockets.websocket_signals_background_job import WebsocketSignalsBackgroundJob
from neptune.new.run import Run
from neptune.new.types.series.string_series import StringSeries
from neptune.new.version import version as parsed_version

__version__ = str(parsed_version)

_logger = logging.getLogger(__name__)


LEGACY_KWARGS = ('project_qualified_name', 'backend')


class RunMode(str, Enum):
    OFFLINE = "offline"
    DEBUG = "debug"
    ASYNC = "async"
    SYNC = "sync"
    READ_ONLY = "read-only"

    def __repr__(self):
        return f'"{self.value}"'


def _check_for_extra_kwargs(caller_name, kwargs: dict):
    for name in LEGACY_KWARGS:
        if name in kwargs:
            raise NeptunePossibleLegacyUsageException()
    if kwargs:
        first_key = next(iter(kwargs.keys()))
        raise TypeError(f"{caller_name}() got an unexpected keyword argument '{first_key}'")


def init(project: Optional[str] = None,
         api_token: Optional[str] = None,
         run: Optional[str] = None,
         custom_run_id: Optional[str] = None,
         mode: str = RunMode.ASYNC,
         name: Optional[str] = None,
         description: Optional[str] = None,
         tags: Optional[Union[List[str], str]] = None,
         source_files: Optional[Union[List[str], str]] = None,
         capture_stdout: bool = True,
         capture_stderr: bool = True,
         capture_hardware_metrics: bool = True,
         fail_on_exception: bool = True,
         monitoring_namespace: Optional[str] = None,
         flush_period: float = 5,
         proxies: Optional[dict] = None,
         capture_traceback: bool = True,
         **kwargs) -> Run:
    """Starts a new tracked run, and append it to the top of the Runs table view.

    Args:
        project(str, optional): Name of a project in a form of `namespace/project_name`. Defaults to `None`.
            If `None`, the value of `NEPTUNE_PROJECT` environment variable will be taken.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
        run (str, optional): An existing run's identifier like 'SAN-1' in case of resuming a tracked run.
            Defaults to `None`.
            A run with such identifier must exist. If `None` is passed, starts a new tracked run.
        custom_run_id (str, optional): A unique identifier to be used when running Neptune in pipelines.
            Defaults to `None`.
            Make sure you are using the same identifier throughout the whole pipeline execution.
        mode (str, optional): Connection mode in which the tracking will work. Defaults to `'async'`.
            Possible values 'async', 'sync', 'offline', 'read-only' and 'debug'.
        name (str, optional): Editable name of the run. Defaults to `'Untitled'`.
            Name is displayed in the run's Details and in Runs table as a column.
        description (str, optional): Editable description of the run. Defaults to `''`.
            Description is displayed in the run's Details and can be displayed in the runs view as a column.
        tags (list of str or str, optional): Tags of the run. Defaults to `[]`.
            They are editable after run is created.
            Tags are displayed in the run's Details and can be viewed in Runs table view as a column.
        source_files (list of str or str, optional): List of source files to be uploaded.
            Uploaded sources are displayed in the run’s Source code tab.
            Unix style pathname pattern expansion is supported. For example, you can pass '*.py' to upload all python
            source files from the current directory.
            If `None` is passed, Python file from which run was created will be uploaded.
        capture_stdout (bool, optional): Whether to send run's stdout. Defaults to `True`.
            Tracked metadata will be stored inside `monitoring_namespace`.
        capture_stderr (bool, optional):  Whether to send run’s stderr. Defaults to `True`.
            Tracked metadata will be stored inside `monitoring_namespace`.
        capture_hardware_metrics (bool, optional): Whether to send hardware monitoring logs
            (CPU, GPU, Memory utilization). Defaults to `True`.
            Tracked metadata will be stored inside `monitoring_namespace`.
        fail_on_exception (bool, optional): Whether to register an uncaught exception handler to this process and,
            in case of an exception, set run's sys/failed to True. Exception is always logged
        monitoring_namespace (str, optional): Namespace inside which all monitoring logs be stored.
            Defaults to 'monitoring'.
        flush_period (float, optional): In an asynchronous (default) connection mode how often asynchronous thread
            should synchronize data with Neptune servers. Defaults to 5.
        proxies (dict of str, optional): Argument passed to HTTP calls made via the Requests library.
            For more information see
            `their proxies section <https://2.python-requests.org/en/master/user/advanced/#proxies>`_.
        capture_traceback (bool, optional):  Whether to send run’s traceback in case of an exception.
            Defaults to `True`.
            Tracked metadata will be stored inside `monitoring/traceback`.

    Returns:
        ``Run``: object that is used to manage the tracked run and log metadata to it.

    Examples:

        >>> import neptune.new as neptune

        >>> # minimal invoke
        ... run = neptune.init()

        >>> # create a tracked run with a name
        ... run = neptune.init(name='first-pytorch-ever')

        >>> # create a tracked run with a name and a description, and no sources files uploaded
        >>> run = neptune.init(name='neural-net-mnist',
        ...                    description='neural net trained on MNIST',
        ...                    source_files=[])

        >>> # Send all py files in cwd (excluding hidden files with names beginning with a dot)
        ... run = neptune.init(source_files='*.py')

        >>> # Send all py files from all subdirectories (excluding hidden files with names beginning with a dot)
        ... # Supported on Python 3.5 and later.
        ... run = neptune.init(source_files='**/*.py')

        >>> # Send all files and directories in cwd (excluding hidden files with names beginning with a dot)
        ... run = neptune.init(source_files='*')

        >>> # Send all files and directories in cwd including hidden files
        ... run = neptune.init(source_files=['*', '.*'])

        >>> # Send files with names being a single character followed by '.py' extension.
        ... run = neptune.init(source_files='?.py')

        >>> # larger example
        ... run = neptune.init(name='first-pytorch-ever',
        ...               description='write longer description here',
        ...               tags=['list-of', 'tags', 'goes-here', 'as-list-of-strings'],
        ...               source_files=['training_with_pytorch.py', 'net.py'])

    You may also want to check `init docs page`_.

    .. _init docs page:
       https://docs.neptune.ai/api-reference/neptune#init
    """
    _check_for_extra_kwargs(init.__name__, kwargs)
    verify_type("project", project, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))
    verify_type("run", run, (str, type(None)))
    verify_type("custom_run_id", custom_run_id, (str, type(None)))
    verify_type("mode", mode, str)
    verify_type("name", name, (str, type(None)))
    verify_type("description", description, (str, type(None)))
    verify_type("capture_stdout", capture_stdout, bool)
    verify_type("capture_stderr", capture_stderr, bool)
    verify_type("capture_hardware_metrics", capture_hardware_metrics, bool)
    verify_type("monitoring_namespace", monitoring_namespace, (str, type(None)))
    verify_type("flush_period", flush_period, (int, float))
    verify_type("proxies", proxies, (dict, type(None)))
    verify_type("capture_traceback", capture_hardware_metrics, bool)
    if tags is not None:
        if isinstance(tags, str):
            tags = [tags]
        else:
            verify_collection_type("tags", tags, str)
    if source_files is not None:
        if isinstance(source_files, str):
            source_files = [source_files]
        else:
            verify_collection_type("source_files", source_files, str)

    name = "Untitled" if run is None and name is None else name
    description = "" if run is None and description is None else description
    hostname = get_hostname() if run is None else None
    custom_run_id = custom_run_id or os.getenv(CUSTOM_RUN_ID_ENV_NAME)
    monitoring_namespace = monitoring_namespace or os.getenv(MONITORING_NAMESPACE) or 'monitoring'

    if run and custom_run_id:
        raise NeptuneRunResumeAndCustomIdCollision()

    if mode == RunMode.ASYNC:
        # TODO Initialize backend in async thread
        backend = HostedNeptuneBackend(
            credentials=Credentials.from_token(api_token=api_token),
            proxies=proxies)
    elif mode == RunMode.SYNC:
        backend = HostedNeptuneBackend(
            credentials=Credentials.from_token(api_token=api_token),
            proxies=proxies)
    elif mode == RunMode.DEBUG:
        backend = NeptuneBackendMock()
    elif mode == RunMode.OFFLINE:
        backend = OfflineNeptuneBackend()
    elif mode == RunMode.READ_ONLY:
        backend = HostedNeptuneBackend(
            credentials=Credentials.from_token(api_token=api_token),
            proxies=proxies)
    else:
        raise ValueError(f'mode should be one of {[m for m in RunMode]}')

    if mode == RunMode.OFFLINE or mode == RunMode.DEBUG:
        project = 'offline/project-placeholder'

    project_obj = project_name_lookup(backend, project)
    project = f'{project_obj.workspace}/{project_obj.name}'

    if run:
        api_run = backend.get_run(project + '/' + run)
    else:
        if mode == RunMode.READ_ONLY:
            raise NeedExistingRunForReadOnlyMode()
        git_ref = get_git_info(discover_git_repo_location())
        if custom_run_id and len(custom_run_id) > 32:
            _logger.warning('Given custom_run_id exceeds 32 characters and it will be ignored.')
            custom_run_id = None

        notebook_id, checkpoint_id = _create_notebook_checkpoint(backend)

        api_run = backend.create_run(project_obj.id, git_ref, custom_run_id, notebook_id, checkpoint_id)

    run_lock = threading.RLock()

    if mode == RunMode.ASYNC:
        run_path = "{}/{}/{}".format(NEPTUNE_RUNS_DIRECTORY, ASYNC_DIRECTORY, api_run.id)
        try:
            execution_id = len(os.listdir(run_path))
        except FileNotFoundError:
            execution_id = 0
        execution_path = "{}/exec-{}-{}".format(run_path, execution_id, datetime.now())
        execution_path = execution_path.replace(" ", "_").replace(":", ".")
        operation_processor = AsyncOperationProcessor(
            api_run.id,
            DiskQueue(Path(execution_path), lambda x: x.to_dict(), Operation.from_dict, run_lock),
            backend,
            run_lock,
            sleep_time=flush_period)
    elif mode == RunMode.SYNC:
        operation_processor = SyncOperationProcessor(api_run.id, backend)
    elif mode == RunMode.DEBUG:
        operation_processor = SyncOperationProcessor(api_run.id, backend)
    elif mode == RunMode.OFFLINE:
        # Run was returned by mocked backend and has some random UUID.
        run_path = "{}/{}/{}".format(NEPTUNE_RUNS_DIRECTORY, OFFLINE_DIRECTORY, api_run.id)
        storage_queue = DiskQueue(Path(run_path),
                                  lambda x: x.to_dict(),
                                  Operation.from_dict,
                                  run_lock)
        operation_processor = OfflineOperationProcessor(storage_queue)
    elif mode == RunMode.READ_ONLY:
        operation_processor = ReadOnlyOperationProcessor(api_run.id, backend)
    else:
        raise ValueError(f'mode should be one of {[m for m in RunMode]}')

    stdout_path = "{}/stdout".format(monitoring_namespace)
    stderr_path = "{}/stderr".format(monitoring_namespace)
    traceback_path = "{}/traceback".format(monitoring_namespace)

    background_jobs = []
    if mode != RunMode.READ_ONLY:
        if capture_stdout:
            background_jobs.append(StdoutCaptureBackgroundJob(attribute_name=stdout_path))
        if capture_stderr:
            background_jobs.append(StderrCaptureBackgroundJob(attribute_name=stderr_path))
        if capture_hardware_metrics:
            background_jobs.append(HardwareMetricReportingJob(attribute_namespace=monitoring_namespace))
        websockets_factory = backend.websockets_factory(project_obj.id, api_run.id)
        if websockets_factory:
            background_jobs.append(WebsocketSignalsBackgroundJob(websockets_factory))
        if capture_traceback:
            background_jobs.append(TracebackJob(traceback_path, fail_on_exception))
        background_jobs.append(PingBackgroundJob())

    _run = Run(api_run.id, backend, operation_processor, BackgroundJobList(background_jobs), run_lock,
               api_run.workspace, api_run.project_name, api_run.short_id, project_obj.id, monitoring_namespace)
    if mode != RunMode.OFFLINE:
        _run.sync(wait=False)

    if mode != RunMode.READ_ONLY:
        if name is not None:
            _run[attr_consts.SYSTEM_NAME_ATTRIBUTE_PATH] = name
        if description is not None:
            _run[attr_consts.SYSTEM_DESCRIPTION_ATTRIBUTE_PATH] = description
        if hostname is not None:
            _run[attr_consts.SYSTEM_HOSTNAME_ATTRIBUTE_PATH] = hostname
        if tags is not None:
            _run[attr_consts.SYSTEM_TAGS_ATTRIBUTE_PATH].add(tags)
        if run is None:
            _run[attr_consts.SYSTEM_FAILED_ATTRIBUTE_PATH] = False

        if capture_stdout and not _run.exists(stdout_path):
            _run.define(stdout_path, StringSeries([]))
        if capture_stderr and not _run.exists(stderr_path):
            _run.define(stderr_path, StringSeries([]))

        if run is None or source_files is not None:
            # upload default sources ONLY if creating a new run
            upload_source_code(source_files=source_files, run=_run)

    _run.start()

    if mode != RunMode.DEBUG:
        click.echo(_run.get_run_url())

        if in_interactive() or in_notebook():
            click.echo(
                "Remember to stop your run once you’ve finished logging your metadata"
                " (https://docs.neptune.ai/api-reference/run#stop)."
                " It will be stopped automatically only when the notebook"
                " kernel/interactive console is terminated.")

    uncaught_exception_handler.activate()

    return _run


def _create_notebook_checkpoint(backend: NeptuneBackend) -> (str, str):
    notebook_id = None
    if os.getenv(NEPTUNE_NOTEBOOK_ID, None) is not None:
        notebook_id = os.environ[NEPTUNE_NOTEBOOK_ID]

    notebook_path = None
    if os.getenv(NEPTUNE_NOTEBOOK_PATH, None) is not None:
        notebook_path = os.environ[NEPTUNE_NOTEBOOK_PATH]

    checkpoint_id = None
    if notebook_id is not None and notebook_path is not None:
        checkpoint_id = create_checkpoint(backend=backend,
                                          notebook_id=notebook_id,
                                          notebook_path=notebook_path)
    return notebook_id, checkpoint_id



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path
from typing import Optional


class SyncOffsetFile:
    def __init__(self, path: Path, default: int = None):
        mode = 'r+' if path.exists() else 'w+'
        self._file = open(path, mode)
        self._default = default
        self._last = self.read()

    def write(self, offset: int) -> None:
        self._file.seek(0)
        self._file.write(str(offset))
        self._file.truncate()
        self._file.flush()
        self._last = offset

    def read(self) -> Optional[int]:
        self._file.seek(0)
        content = self._file.read()
        if not content:
            return self._default
        return int(content)

    def read_local(self) -> Optional[int]:
        return self._last

    def flush(self):
        self._file.flush()

    def close(self):
        self._file.close()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time

import jwt

from tests.neptune.random_utils import a_string

SECRET = 'secret'


def an_access_token():
    return jwt.encode({'exp': time.time(), 'azp': a_string(), 'iss': 'http://{}.com'.format(a_string())}, SECRET)


def a_refresh_token():
    return jwt.encode({'exp': 0, 'azp': a_string(), 'iss': 'http://{}.com'.format(a_string())}, SECRET)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import uuid

from mock import MagicMock, patch

from neptune.new.exceptions import ArtifactUploadingError
from neptune.new.internal.backends.api_model import ArtifactModel
from neptune.new.internal.artifacts.types import ArtifactFileData
from neptune.new.internal.backends.hosted_artifact_operations import (
    track_to_new_artifact,
    track_to_existing_artifact
)


class TestHostedArtifactOperations(unittest.TestCase):
    def setUp(self) -> None:
        self.artifact_hash = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        self.files = [
            ArtifactFileData(
                "fname.txt",
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "test",
                {}
            ),
        ]
        self.project_id = str(uuid.uuid4())
        self.parent_identifier = str(uuid.uuid4())

    @patch('neptune.new.internal.backends.hosted_artifact_operations._compute_artifact_hash')
    @patch('neptune.new.internal.backends.hosted_artifact_operations._extract_file_list')
    @patch('neptune.new.internal.backends.hosted_artifact_operations.create_new_artifact')
    def test_track_to_new_artifact_calls_creation(
            self,
            create_new_artifact,
            _extract_file_list,
            _compute_artifact_hash
    ):
        # given
        swagger_mock = self._get_swagger_mock()
        _compute_artifact_hash.return_value = self.artifact_hash
        _extract_file_list.return_value = self.files

        # when
        track_to_new_artifact(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            path=["sub", "one"],
            parent_identifier=self.parent_identifier,
            entries=[("/path/to/file", '/path/to')],
            default_request_params={}
        )

        # then
        create_new_artifact.assert_called_once_with(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            artifact_hash=self.artifact_hash,
            parent_identifier=self.parent_identifier,
            size=None,
            default_request_params={}
        )

    @patch('neptune.new.internal.backends.hosted_artifact_operations._compute_artifact_hash')
    @patch('neptune.new.internal.backends.hosted_artifact_operations._extract_file_list')
    @patch('neptune.new.internal.backends.hosted_artifact_operations.create_new_artifact')
    @patch('neptune.new.internal.backends.hosted_artifact_operations.upload_artifact_files_metadata')
    def test_track_to_new_artifact_calls_upload(
            self,
            upload_artifact_files_metadata,
            create_new_artifact,
            _extract_file_list,
            _compute_artifact_hash
    ):
        # given
        swagger_mock = self._get_swagger_mock()
        _compute_artifact_hash.return_value = self.artifact_hash
        _extract_file_list.return_value = self.files
        create_new_artifact.return_value = ArtifactModel(
            received_metadata=False,
            hash=self.artifact_hash,
            size=len(self.files)
        )

        # when
        track_to_new_artifact(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            path=["sub", "one"],
            parent_identifier=self.parent_identifier,
            entries=[("/path/to/file", '/path/to')],
            default_request_params={}
        )

        # then
        upload_artifact_files_metadata.assert_called_once_with(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            artifact_hash=self.artifact_hash,
            files=self.files,
            default_request_params={}
        )

    @patch('neptune.new.internal.backends.hosted_artifact_operations._compute_artifact_hash')
    @patch('neptune.new.internal.backends.hosted_artifact_operations._extract_file_list')
    def test_track_to_new_artifact_raises_exception(
            self,
            _extract_file_list,
            _compute_artifact_hash
    ):
        # given
        swagger_mock = self._get_swagger_mock()
        _compute_artifact_hash.return_value = self.artifact_hash
        _extract_file_list.return_value = []

        # when
        with self.assertRaises(ArtifactUploadingError):
            track_to_new_artifact(
                swagger_client=swagger_mock,
                project_id=self.project_id,
                path=["sub", "one"],
                parent_identifier=self.parent_identifier,
                entries=[("/path/to/file", '/path/to')],
                default_request_params={}
            )

    @patch('neptune.new.internal.backends.hosted_artifact_operations._compute_artifact_hash')
    @patch('neptune.new.internal.backends.hosted_artifact_operations._extract_file_list')
    @patch('neptune.new.internal.backends.hosted_artifact_operations.create_artifact_version')
    def test_track_to_existing_artifact_calls_version(
            self,
            create_artifact_version,
            _extract_file_list,
            _compute_artifact_hash
    ):
        # given
        swagger_mock = self._get_swagger_mock()
        _extract_file_list.return_value = self.files

        # when
        track_to_existing_artifact(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            path=["sub", "one"],
            artifact_hash=self.artifact_hash,
            parent_identifier=self.parent_identifier,
            entries=[("/path/to/file", '/path/to')],
            default_request_params={}
        )

        # then
        create_artifact_version.assert_called_once_with(
            swagger_client=swagger_mock,
            project_id=self.project_id,
            artifact_hash=self.artifact_hash,
            parent_identifier=self.parent_identifier,
            files=self.files,
            default_request_params={}
        )

    @patch('neptune.new.internal.backends.hosted_artifact_operations._compute_artifact_hash')
    @patch('neptune.new.internal.backends.hosted_artifact_operations._extract_file_list')
    def test_track_to_existing_artifact_raises_exception(
            self,
            _extract_file_list,
            _compute_artifact_hash
    ):
        # given
        swagger_mock = self._get_swagger_mock()
        _compute_artifact_hash.return_value = self.artifact_hash
        _extract_file_list.return_value = []

        # when
        with self.assertRaises(ArtifactUploadingError):
            track_to_existing_artifact(
                swagger_client=swagger_mock,
                project_id=self.project_id,
                path=["sub", "one"],
                artifact_hash='abcdef',
                parent_identifier=self.parent_identifier,
                entries=[("/path/to/file", '/path/to')],
                default_request_params={}
            )

    @staticmethod
    def _get_swagger_mock():
        swagger_mock = MagicMock()
        swagger_mock.swagger_spec.http_client = MagicMock()
        swagger_mock.swagger_spec.api_url = "ui.neptune.ai"
        swagger_mock.api.createNewArtifact.operation.path_name = "/createNewArtifact"
        swagger_mock.api.uploadArtifactFilesMetadata.operation.path_name = "/uploadArtifactFilesMetadata"
        return swagger_mock



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

try:
    import psutil
    PSUTIL_INSTALLED = True
except ImportError:
    PSUTIL_INSTALLED = False


class CustomAbortImpl(object):
    def __init__(self, runnable):
        self.__runnable = runnable

    def abort(self):
        self.__runnable()


class DefaultAbortImpl(object):
    KILL_TIMEOUT = 5

    def __init__(self, pid):
        self._pid = pid

    @staticmethod
    def requirements_installed():
        return PSUTIL_INSTALLED

    def abort(self):
        try:
            processes = self._get_process_with_children(psutil.Process(self._pid))
        except psutil.NoSuchProcess:
            processes = []

        for p in processes:
            self._abort(p)
        _, alive = psutil.wait_procs(processes, timeout=self.KILL_TIMEOUT)
        for p in alive:
            self._kill(p)

    @staticmethod
    def _get_process_with_children(process):
        try:
            return [process] + process.children(recursive=True)
        except psutil.NoSuchProcess:
            return []

    @staticmethod
    def _abort(process):
        try:
            process.terminate()
        except psutil.NoSuchProcess:
            pass

    def _kill(self, process):
        for process in self._get_process_with_children(process):
            try:
                if process.is_running():
                    process.kill()
            except psutil.NoSuchProcess:
                pass



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import platform

from neptune import envs

UNIX_STYLES = {'h1': '\033[95m',
               'h2': '\033[94m',
               'blue': '\033[94m',
               'python': '\033[96m',
               'bash': '\033[95m',
               'warning': '\033[93m',
               'correct': '\033[92m',
               'fail': '\033[91m',
               'bold': '\033[1m',
               'underline': '\033[4m',
               'end': '\033[0m'}

WINDOWS_STYLES = {'h1': '',
                  'h2': '',
                  'python': '',
                  'bash': '',
                  'warning': '',
                  'correct': '',
                  'fail': '',
                  'bold': '',
                  'underline': '',
                  'end': ''}

EMPTY_STYLES = {'h1': '',
                'h2': '',
                'python': '',
                'bash': '',
                'warning': '',
                'correct': '',
                'fail': '',
                'bold': '',
                'underline': '',
                'end': ''}

if platform.system() in ['Linux', 'Darwin']:
    STYLES = UNIX_STYLES
elif platform.system() == 'Windows':
    STYLES = WINDOWS_STYLES
else:
    STYLES = EMPTY_STYLES


class NeptuneException(Exception):
    pass


class NeptuneUninitializedException(NeptuneException):
    def __init__(self):
        message = """
{h1}     
----NeptuneUninitializedException---------------------------------------------------------------------------------------
{end}
You must initialize neptune-client before you create an experiment.

Looks like you forgot to add:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}
    
before you ran:
    {python}neptune.create_experiment(){end}
    
You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html
    
{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
""".format(**STYLES)
        super(NeptuneUninitializedException, self).__init__(message)


class FileNotFound(NeptuneException):
    def __init__(self, path):
        super(FileNotFound, self).__init__("File {} doesn't exist.".format(path))


class NotAFile(NeptuneException):
    def __init__(self, path):
        super(NotAFile, self).__init__("Path {} is not a file.".format(path))


class NotADirectory(NeptuneException):
    def __init__(self, path):
        super(NotADirectory, self).__init__("Path {} is not a directory.".format(path))


class InvalidNotebookPath(NeptuneException):
    def __init__(self, path):
        super(InvalidNotebookPath, self).__init__(
            "File {} is not a valid notebook. Should end with .ipynb.".format(path))


class NoChannelValue(NeptuneException):
    def __init__(self):
        super(NoChannelValue, self).__init__('No channel value provided.')


class NeptuneLibraryNotInstalledException(NeptuneException):
    def __init__(self, library):
        message = """
{h1}     
----NeptuneLibraryNotInstalledException---------------------------------------------------------------------------------
{end}
Looks like library {library} wasn't installed.

To install run:
    {bash}pip install {library}{end}

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/getting-started/installation/index.html
    
{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneLibraryNotInstalledException, self).__init__(message.format(
            library=library,
            **STYLES
        ))


class InvalidChannelValue(NeptuneException):
    def __init__(self, expected_type, actual_type):
        super(InvalidChannelValue, self).__init__(
            'Invalid channel value type. Expected: {expected}, actual: {actual}.'.format(
                expected=expected_type, actual=actual_type))


class NeptuneNoExperimentContextException(NeptuneException):
    def __init__(self):
        message = """
{h1}  
----NeptuneNoExperimentContextException---------------------------------------------------------------------------------
{end}
Neptune couldn't find an active experiment.

Looks like you forgot to run:
    {python}neptune.create_experiment(){end}
    
You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneNoExperimentContextException, self).__init__(message.format(**STYLES))


class NeptuneMissingApiTokenException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneMissingApiTokenException-------------------------------------------------------------------------------------
{end}
Neptune client couldn't find your API token.

Learn how to get it in this docs page:
https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html

There are two options to add it:
    - specify it in your code 
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass the token to {bold}neptune.init(){end} via {bold}api_token{end} argument:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}=YOUR_API_TOKEN{end}
        
    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}=YOUR_API_TOKEN{end}
        
and skip the {bold}api_token{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME'){end}
    
You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneMissingApiTokenException, self).__init__(message.format(
            env_api_token=envs.API_TOKEN_ENV_NAME,
            **STYLES
        ))


class NeptuneMissingProjectQualifiedNameException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneMissingProjectQualifiedNameException-------------------------------------------------------------------------
{end}
Neptune client couldn't find your project name.

There are two options two add it:
    - specify it in your code 
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass it to {bold}neptune.init(){end} via {bold}project_qualified_name{end} argument:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
       {bash}export {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

    {correct}Windows{end}
    In your CMD run:
       {bash}set {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

and skip the {bold}project_qualified_name{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(api_token='YOUR_API_TOKEN'){end}

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/workspace-project-and-user-management/index.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html
    
{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneMissingProjectQualifiedNameException, self).__init__(message.format(
            env_project=envs.PROJECT_ENV_NAME,
            **STYLES
        ))


class NeptuneIncorrectProjectQualifiedNameException(NeptuneException):
    def __init__(self, project_qualified_name):
        message = """
{h1}
----NeptuneIncorrectProjectQualifiedNameException-----------------------------------------------------------------------
{end}
Project qualified name {fail}"{project_qualified_name}"{end} you specified was incorrect.

The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
It has two parts:
    - {correct}WORKSPACE{end}: which can be your username or your organization name
    - {correct}PROJECT_NAME{end}: which is the actual project name you chose 

For example, a project {correct}neptune-ai/credit-default-prediction{end} parts are:
    - {correct}neptune-ai{end}: {underline}WORKSPACE{end} our company organization name
    - {correct}credit-default-prediction{end}: {underline}PROJECT_NAME{end} a project name
    
The URL to this project looks like this: https://app.neptune.ai/neptune-ai/credit-default-prediction

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/workspace-project-and-user-management/index.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneIncorrectProjectQualifiedNameException, self).__init__(message.format(
            project_qualified_name=project_qualified_name,
            **STYLES
        ))


class InvalidNeptuneBackend(NeptuneException):
    def __init__(self, provided_backend_name):
        super(InvalidNeptuneBackend, self).__init__(
            'Unknown {} "{}". '
            'Use this environment variable to modify neptune-client behaviour at runtime, '
            'e.g. using {}=offline allows you to run your code without logging anything to Neptune'
            ''.format(envs.BACKEND, provided_backend_name, envs.BACKEND))


class DeprecatedApiToken(NeptuneException):
    def __init__(self, app_url):
        super(DeprecatedApiToken, self).__init__(
            "Your API token is deprecated. Please visit {} to get a new one.".format(app_url))


class CannotResolveHostname(NeptuneException):
    def __init__(self, host):
        message = """
{h1}
----CannotResolveHostname-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to resolve hostname {host}.

What should I do?
    - Check if your computer is connected to the internet.
    - Check if your computer should use any proxy to access internet.
      If so, you may want to use {python}proxies{end} parameter of {python}neptune.init(){end} function.
      See https://docs-legacy.neptune.ai/api-reference/neptune/index.html#neptune.init
      and https://requests.readthedocs.io/en/master/user/advanced/#proxies
"""
        super(CannotResolveHostname, self).__init__(message.format(
            host=host,
            **STYLES
        ))


class UnsupportedClientVersion(NeptuneException):
    def __init__(self, version, minVersion, maxVersion):
        super(UnsupportedClientVersion, self).__init__(
            "This client version ({}) is not supported. Please install neptune-client{}".format(
                version,
                "==" + str(maxVersion) if maxVersion else ">=" + str(minVersion)
            ))


class UnsupportedInAlphaException(NeptuneException):
    """Raised for operations which was available in old client,
    but aren't supported in alpha version"""


class DownloadSourcesException(UnsupportedInAlphaException):
    message = """
{h1}
----DownloadSourcesException-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to download single file from sources.

Why am I seeing this?
    Your project "{project}" has been migrated to new structure.
    Old version of `neptune-api` is not supporting downloading particular source files.
    We recommend you to use new version of api: `neptune.new`.
    {correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help

If you don't want to adapt your code to new api yet,
you can use `download_sources` with `path` parameter set to None.
"""

    def __init__(self, experiment):
        assert self.message is not None
        super().__init__(
            self.message.format(
                project=experiment._project.internal_id,
                **STYLES,
            )
        )


class DownloadArtifactsUnsupportedException(UnsupportedInAlphaException):
    message = """
{h1}
----DownloadArtifactsUnsupportedException-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to download artifacts.
Function `download_artifacts` is deprecated.

Why am I seeing this?
    Your project "{project}" has been migrated to new structure.
    Old version of `neptune-api` is not supporting downloading artifact directories.
    We recommend you to use new version of api: `neptune.new`.
    {correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help

If you don't want to adapt your code to new api yet,
you can use `download_artifact` and download files one by one.
"""

    def __init__(self, experiment):
        assert self.message is not None
        super().__init__(
            self.message.format(
                project=experiment._project.internal_id,
                **STYLES,
            )
        )


class DownloadArtifactUnsupportedException(UnsupportedInAlphaException):
    message = """
{h1}
----DownloadArtifactUnsupportedException-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to download attribute: "{artifact_path}".
It's not present in experiment {experiment} or is a directory.

Why am I seeing this?
    Your project "{project}" has been migrated to new structure.
    Old version of `neptune-api` is not supporting downloading whole artifact directories.
    We recommend you to use new version of api: `neptune.new`.
    {correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help

If you don't want to adapt your code to new api yet:
    - Make sure that artifact "{artifact_path}" is present in experiment "{experiment}".
    - Make sure that you're addressing artifact which is file, not whole directory.
"""

    def __init__(self, artifact_path, experiment):
        assert self.message is not None
        super().__init__(
            self.message.format(
                artifact_path=artifact_path,
                experiment=experiment.id,
                project=experiment._project.internal_id,
                **STYLES,
            )
        )


class DeleteArtifactUnsupportedInAlphaException(UnsupportedInAlphaException):
    message = """
{h1}
----DeleteArtifactUnsupportedInAlphaException-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to delete attribute: "{artifact_path}".
It's not present in experiment {experiment} or is a directory.

Why am I seeing this?
    Your project "{project}" has been migrated to new structure.
    Old version of `neptune-api` is not supporting deleting whole artifact directories.
    We recommend you to use new version of api: `neptune.new`.
    {correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help

If you don't want to adapt your code to new api yet:
    - Make sure that artifact "{artifact_path}" is present in experiment "{experiment}".
    - Make sure that you're addressing artifact which is file, not whole directory.
"""

    def __init__(self, artifact_path, experiment):
        assert self.message is not None
        super().__init__(
            self.message.format(
                artifact_path=artifact_path,
                experiment=experiment.id,
                project=experiment._project.internal_id,
                **STYLES,
            )
        )


class NeptuneIncorrectImportException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneIncorrectImportException----------------------------------------------------------------
{end}
It seems you are trying to use the new Python API, but imported the legacy API.

Simply update your import statement to:
    {python}import neptune.new as neptune{end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/migration-guide

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class ProjectMigratedToNewStructure(NeptuneException):
    pass



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=protected-access
import uuid
import pathlib
import tempfile
from unittest.mock import Mock
from mock import MagicMock, call

from _pytest.monkeypatch import MonkeyPatch

from neptune.new.attributes.atoms.artifact import Artifact
from neptune.new.exceptions import NeptuneUnhandledArtifactTypeException
from neptune.new.internal.artifacts.types import ArtifactFileData, ArtifactDriver, ArtifactDriversMap
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.types.atoms.artifact import Artifact as ArtifactAttr
from neptune.new.internal.operation import TrackFilesToArtifact

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestArtifact(TestAttributeBase):
    def setUp(self):
        self.monkeypatch = MonkeyPatch()

        self.wait = self._random_wait()
        self.op_processor = MagicMock()
        self.exp = self._create_run(processor=self.op_processor)
        self.path = self._random_path()
        self.path_str = path_to_str(self.path)

        self.artifact_hash = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
        self.artifact_files = [
            ArtifactFileData(
                file_path="fname.txt",
                file_hash="e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                type="test",
                size=213,
                metadata={},
            ),
            ArtifactFileData(
                file_path="subdir/other.mp3",
                file_hash="e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                type="test",
                metadata={}
            ),
        ]

        self.exp.set_attribute(self.path_str, Artifact(self.exp, self.path))
        self.exp._backend._runs[self.exp._id].set(self.path, ArtifactAttr(self.artifact_hash))
        self.exp._backend._artifacts[self.exp._project_id, self.artifact_hash] = self.artifact_files

        self._downloads = set()

        class TestArtifactDriver(ArtifactDriver):
            @classmethod
            def get_type(cls):
                return "test"

            @classmethod
            def matches(cls, path: str) -> bool:
                return False

            @classmethod
            def get_tracked_files(cls, path, destination=None):
                return []

            @classmethod
            def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
                destination.touch()

        self.test_artifact_driver = TestArtifactDriver

    def tearDown(self):
        self.monkeypatch.undo()

    def test_fetch_hash(self):
        fetched_hash = self.exp[self.path_str].fetch_hash()
        self.assertEqual(self.artifact_hash, fetched_hash)

    def test_fetch_files_list(self):
        fetched_hash = self.exp[self.path_str].fetch_files_list()
        self.assertEqual(self.artifact_files, fetched_hash)

    def test_download(self):
        self.monkeypatch.setattr(ArtifactDriversMap, 'match_type', Mock(return_value=self.test_artifact_driver))

        with tempfile.TemporaryDirectory() as temporary:
            self.exp[self.path_str].download(temporary)
            temporary_path = pathlib.Path(temporary)

            self.assertTrue((temporary_path / "fname.txt").exists())
            self.assertTrue((temporary_path / "subdir" / "other.mp3").exists())

    def test_download_unknown_type(self):
        self.monkeypatch.setattr(
            ArtifactDriversMap, "match_type", Mock(side_effect=NeptuneUnhandledArtifactTypeException("test")))

        with tempfile.TemporaryDirectory() as temporary:
            with self.assertRaises(NeptuneUnhandledArtifactTypeException):
                self.exp[self.path_str].download(temporary)

            temporary_path = pathlib.Path(temporary)
            contents = list(temporary_path.iterdir())
            self.assertListEqual(contents, [])

    def test_track_files_to_artifact(self):
        source_location = str(uuid.uuid4())
        destination = str(uuid.uuid4())
        source_location2 = str(uuid.uuid4())
        destination2 = str(uuid.uuid4())

        var = Artifact(self.exp, self.path)
        var.track_files(
            path=source_location,
            destination=destination,
            wait=self.wait
        )
        var.track_files(
            path=source_location2,
            destination=destination2,
            wait=self.wait
        )

        self.op_processor.enqueue_operation.assert_has_calls([
            call(
                TrackFilesToArtifact(self.path, self.exp._project_id, [(source_location, destination)]),
                self.wait
            ),
            call(
                TrackFilesToArtifact(
                    self.path,
                    self.exp._project_id,
                    [(source_location2, destination2)]
                ),
                self.wait
            )
        ])



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from typing import TypeVar, Generic

from neptune.new.types.atoms import GitRef
from neptune.new.types.atoms.artifact import Artifact
from neptune.new.types.atoms.boolean import Boolean
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.integer import Integer
from neptune.new.types.atoms.string import String
from neptune.new.types.atoms.file import File
from neptune.new.types.file_set import FileSet
from neptune.new.types.namespace import Namespace
from neptune.new.types.series.float_series import FloatSeries
from neptune.new.types.series.file_series import FileSeries
from neptune.new.types.series.string_series import StringSeries
from neptune.new.types.sets.string_set import StringSet
from neptune.new.types.value import Value

Ret = TypeVar('Ret')


class ValueVisitor(Generic[Ret]):

    def visit(self, value: Value) -> Ret:
        return value.accept(self)

    @abc.abstractmethod
    def visit_float(self, value: Float) -> Ret:
        pass

    @abc.abstractmethod
    def visit_integer(self, value: Integer) -> Ret:
        pass

    @abc.abstractmethod
    def visit_boolean(self, value: Boolean) -> Ret:
        pass

    @abc.abstractmethod
    def visit_string(self, value: String) -> Ret:
        pass

    @abc.abstractmethod
    def visit_datetime(self, value: Datetime) -> Ret:
        pass

    @abc.abstractmethod
    def visit_artifact(self, value: Artifact) -> Ret:
        pass

    @abc.abstractmethod
    def visit_file(self, value: File) -> Ret:
        pass

    @abc.abstractmethod
    def visit_file_set(self, value: FileSet) -> Ret:
        pass

    @abc.abstractmethod
    def visit_float_series(self, value: FloatSeries) -> Ret:
        pass

    @abc.abstractmethod
    def visit_string_series(self, value: StringSeries) -> Ret:
        pass

    @abc.abstractmethod
    def visit_image_series(self, value: FileSeries) -> Ret:
        pass

    @abc.abstractmethod
    def visit_string_set(self, value: StringSet) -> Ret:
        pass

    @abc.abstractmethod
    def visit_git_ref(self, value: GitRef) -> Ret:
        pass

    @abc.abstractmethod
    def visit_namespace(self, value: Namespace) -> Ret:
        pass



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.new.internal.backends.api_model import AttributeType


class NoValue:
    pass


VALUE = "value"
LAST_VALUE = "last"
VALUES = "values"

atomic_attribute_types_map = {
    AttributeType.FLOAT.value: "floatProperties",
    AttributeType.INT.value: "intProperties",
    AttributeType.BOOL.value: "boolProperties",
    AttributeType.STRING.value: "stringProperties",
    AttributeType.DATETIME.value: "datetimeProperties",
    AttributeType.RUN_STATE.value: "experimentStateProperties",
    AttributeType.NOTEBOOK_REF.value: "notebookRefProperties",
}

value_series_attribute_types_map = {
    AttributeType.FLOAT_SERIES.value: "floatSeriesProperties",
    AttributeType.STRING_SERIES.value: "stringSeriesProperties",
}

value_set_attribute_types_map = {
    AttributeType.STRING_SET.value: "stringSetProperties",
}

# TODO: nicer mapping?
_unmapped_attribute_types_map = {
    AttributeType.FILE_SET.value: "fileSetProperties",   # TODO: return size?
    AttributeType.FILE.value: "fileProperties",  # TODO: name? size?
    AttributeType.IMAGE_SERIES.value: "imageSeriesProperties",  # TODO: return last step?
    AttributeType.GIT_REF.value: "gitRefProperties",  # TODO: commit? branch?
}


def map_attribute_result_to_value(attribute):
    for attribute_map, value_key in [
            (atomic_attribute_types_map, VALUE),
            (value_series_attribute_types_map, LAST_VALUE),
            (value_set_attribute_types_map, VALUES),
    ]:
        source_property = attribute_map.get(attribute.type)
        if source_property is not None:
            mapped_attribute_entry = getattr(attribute, source_property)
            return getattr(mapped_attribute_entry, value_key)
    return NoValue



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import platform
from typing import Dict, Tuple

from bravado.client import SwaggerClient
from bravado.http_client import HttpClient
from bravado.requests_client import RequestsClient
from packaging import version

from neptune.new.internal.backends.utils import (
    with_api_exceptions_handler,
    build_operation_url,
    update_session_proxies,
    verify_host_resolution,
    create_swagger_client,
    verify_client_version,
    cache,
)
from neptune.new.internal.backends.api_model import ClientConfig
from neptune.new.internal.credentials import Credentials
from neptune.new.version import version as neptune_client_version
from neptune.new.exceptions import UnsupportedClientVersion
# TODO: Do not use NeptuneAuthenticator from old_neptune. Move it to new package.
from neptune.oauth import NeptuneAuthenticator

BACKEND_SWAGGER_PATH = "/api/backend/swagger.json"
LEADERBOARD_SWAGGER_PATH = "/api/leaderboard/swagger.json"
ARTIFACTS_SWAGGER_PATH = "/api/artifacts/swagger.json"

CONNECT_TIMEOUT = 30  # helps detecting internet connection lost
REQUEST_TIMEOUT = None

DEFAULT_REQUEST_KWARGS = {
    '_request_options': {
        "connect_timeout": CONNECT_TIMEOUT,
        "timeout": REQUEST_TIMEOUT,
        "headers": {"X-Neptune-LegacyClient": "false"}
    }
}


def create_http_client(ssl_verify: bool, proxies: Dict[str, str]) -> RequestsClient:
    http_client = RequestsClient(ssl_verify=ssl_verify)
    http_client.session.verify = ssl_verify

    update_session_proxies(http_client.session, proxies)

    user_agent = 'neptune-client/{lib_version} ({system}, python {python_version})'.format(
        lib_version=neptune_client_version,
        system=platform.platform(),
        python_version=platform.python_version())
    http_client.session.headers.update({'User-Agent': user_agent})

    return http_client


@cache
def _get_token_client(
        credentials: Credentials,
        ssl_verify: bool,
        proxies: Dict[str, str],
        endpoint_url: str = None
) -> SwaggerClient:
    config_api_url = credentials.api_url_opt or credentials.token_origin_address
    if proxies is None:
        verify_host_resolution(config_api_url)

    token_http_client = create_http_client(ssl_verify, proxies)

    return create_swagger_client(
        build_operation_url(endpoint_url or config_api_url, BACKEND_SWAGGER_PATH),
        token_http_client
    )


@cache
@with_api_exceptions_handler
def get_client_config(credentials: Credentials, ssl_verify: bool, proxies: Dict[str, str]) -> ClientConfig:
    backend_client = _get_token_client(credentials=credentials, ssl_verify=ssl_verify, proxies=proxies)

    config = backend_client.api.getClientConfig(
        X_Neptune_Api_Token=credentials.api_token,
        alpha="true",
        **DEFAULT_REQUEST_KWARGS,
    ).response().result

    if hasattr(config, "pyLibVersions"):
        min_recommended = getattr(config.pyLibVersions, "minRecommendedVersion", None)
        min_compatible = getattr(config.pyLibVersions, "minCompatibleVersion", None)
        max_compatible = getattr(config.pyLibVersions, "maxCompatibleVersion", None)
    else:
        raise UnsupportedClientVersion(neptune_client_version, max_version="0.4.111")

    return ClientConfig(
        api_url=config.apiUrl,
        display_url=config.applicationUrl,
        min_recommended_version=version.parse(min_recommended) if min_recommended else None,
        min_compatible_version=version.parse(min_compatible) if min_compatible else None,
        max_compatible_version=version.parse(max_compatible) if max_compatible else None
    )


@cache
def create_http_client_with_auth(
        credentials: Credentials,
        ssl_verify: bool,
        proxies: Dict[str, str]
) -> Tuple[RequestsClient, ClientConfig]:
    client_config = get_client_config(
        credentials=credentials,
        ssl_verify=ssl_verify,
        proxies=proxies
    )

    config_api_url = credentials.api_url_opt or credentials.token_origin_address

    verify_client_version(client_config, neptune_client_version)

    endpoint_url = None
    if config_api_url != client_config.api_url:
        endpoint_url = build_operation_url(client_config.api_url, BACKEND_SWAGGER_PATH)

    http_client = create_http_client(ssl_verify=ssl_verify, proxies=proxies)
    http_client.authenticator = NeptuneAuthenticator(
        credentials.api_token,
        _get_token_client(
            credentials=credentials,
            ssl_verify=ssl_verify,
            proxies=proxies,
            endpoint_url=endpoint_url
        ),
        ssl_verify,
        proxies
    )

    return http_client, client_config


@cache
def create_backend_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClient:
    return create_swagger_client(
        build_operation_url(client_config.api_url, BACKEND_SWAGGER_PATH),
        http_client
    )


@cache
def create_leaderboard_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClient:
    return create_swagger_client(
        build_operation_url(client_config.api_url, LEADERBOARD_SWAGGER_PATH),
        http_client
    )


@cache
def create_artifacts_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClient:
    return create_swagger_client(
        build_operation_url(client_config.api_url, ARTIFACTS_SWAGGER_PATH),
        http_client
    )



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=too-many-lines

import io
import json
import logging
import math
import os
import re
import sys
import uuid
from datetime import time
from functools import partial
from http.client import NOT_FOUND  # pylint:disable=no-name-in-module
from io import StringIO
from itertools import groupby
from typing import Dict

import requests
import six
from bravado.exception import HTTPBadRequest, HTTPNotFound, HTTPUnprocessableEntity, HTTPConflict
from neptune.internal.websockets.reconnecting_websocket_factory import ReconnectingWebsocketFactory

from neptune.api_exceptions import (
    ChannelAlreadyExists,
    ChannelNotFound,
    ChannelsValuesSendBatchError,
    ExperimentAlreadyFinished,
    ExperimentLimitReached,
    ExperimentNotFound,
    ExperimentValidationError,
    NotebookNotFound,
    PathInProjectNotFound,
    ProjectNotFound,
)
from neptune.backend import LeaderboardApiClient
from neptune.checkpoint import Checkpoint
from neptune.exceptions import FileNotFound
from neptune.experiments import Experiment
from neptune.internal.api_clients.hosted_api_clients.mixins import HostedNeptuneMixin
from neptune.internal.storage.storage_utils import UploadEntry, normalize_file_name, upload_to_storage
from neptune.internal.utils.http_utils import extract_response_field, handle_quota_limits
from neptune.model import ChannelWithLastValue, LeaderboardEntry
from neptune.notebook import Notebook
from neptune.utils import NoopObject, assure_directory_exists, with_api_exceptions_handler

_logger = logging.getLogger(__name__)


class HostedNeptuneLeaderboardApiClient(HostedNeptuneMixin, LeaderboardApiClient):
    @with_api_exceptions_handler
    def __init__(self, backend_api_client):
        self._backend_api_client = backend_api_client

        self._client_config = self._create_client_config(api_token=self.credentials.api_token,
                                                         backend_client=self.backend_client)

        self.leaderboard_swagger_client = self._get_swagger_client(
            '{}/api/leaderboard/swagger.json'.format(self._client_config.api_url),
            self._backend_api_client.http_client
        )

        if sys.version_info >= (3, 7):
            try:
                # pylint: disable=no-member
                os.register_at_fork(after_in_child=self._handle_fork_in_child)
            except AttributeError:
                pass

    def _handle_fork_in_child(self):
        self.leaderboard_swagger_client = NoopObject()

    @property
    def http_client(self):
        return self._backend_api_client.http_client

    @property
    def backend_client(self):
        return self._backend_api_client.backend_client

    @property
    def authenticator(self):
        return self._backend_api_client.authenticator

    @property
    def credentials(self):
        return self._backend_api_client.credentials

    @property
    def backend_swagger_client(self):
        return self._backend_api_client.backend_swagger_client

    @property
    def client_lib_version(self):
        return self._backend_api_client.client_lib_version

    @property
    def api_address(self):
        return self._client_config.api_url

    @property
    def display_address(self):
        return self._backend_api_client.display_address

    @property
    def proxies(self):
        return self._backend_api_client.proxies

    @with_api_exceptions_handler
    def get_project_members(self, project_identifier):
        try:
            r = self.backend_swagger_client.api.listProjectMembers(projectIdentifier=project_identifier).response()
            return r.result
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier)

    @with_api_exceptions_handler
    def get_leaderboard_entries(self, project,
                                entry_types=None, ids=None,
                                states=None, owners=None, tags=None,
                                min_running_time=None):
        try:
            if entry_types is None:
                entry_types = ['experiment', 'notebook']

            def get_portion(limit, offset):
                return self.leaderboard_swagger_client.api.getLeaderboard(
                    projectIdentifier=project.full_id,
                    entryType=entry_types,
                    shortId=ids, groupShortId=None, state=states, owner=owners, tags=tags,
                    tagsMode='and', minRunningTimeSeconds=min_running_time,
                    sortBy=['shortId'], sortFieldType=['native'], sortDirection=['ascending'],
                    limit=limit, offset=offset
                ).response().result.entries

            return [LeaderboardEntry(e) for e in self._get_all_items(get_portion, step=100)]
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

    def websockets_factory(self, project_id, experiment_id):
        base_url = re.sub(r'^http', 'ws', self.api_address) + '/api/notifications/v1'
        return ReconnectingWebsocketFactory(
            backend=self,
            url=base_url + '/experiments/' + experiment_id + '/operations'
        )

    @with_api_exceptions_handler
    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        try:
            csv = StringIO()
            csv.write(
                self.backend_swagger_client.api.getChannelValuesCSV(
                    experimentId=experiment.internal_id, channelId=channel_internal_id
                ).response().incoming_response.text
            )
            csv.seek(0)
            return csv
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def get_metrics_csv(self, experiment):
        try:
            csv = StringIO()
            csv.write(
                self.backend_swagger_client.api.getSystemMetricsCSV(
                    experimentId=experiment.internal_id
                ).response().incoming_response.text
            )
            csv.seek(0)
            return csv
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def create_experiment(self,
                          project,
                          name,
                          description,
                          params,
                          properties,
                          tags,
                          abortable,
                          monitored,
                          git_info,
                          hostname,
                          entrypoint,
                          notebook_id,
                          checkpoint_id):
        if not isinstance(name, six.string_types):
            raise ValueError("Invalid name {}, should be a string.".format(name))
        if not isinstance(description, six.string_types):
            raise ValueError("Invalid description {}, should be a string.".format(description))
        if not isinstance(params, dict):
            raise ValueError("Invalid params {}, should be a dict.".format(params))
        if not isinstance(properties, dict):
            raise ValueError("Invalid properties {}, should be a dict.".format(properties))
        if not isinstance(hostname, six.string_types):
            raise ValueError("Invalid hostname {}, should be a string.".format(hostname))
        if entrypoint is not None and not isinstance(entrypoint, six.string_types):
            raise ValueError("Invalid entrypoint {}, should be a string.".format(entrypoint))

        ExperimentCreationParams = self.backend_swagger_client.get_model('ExperimentCreationParams')
        GitInfoDTO = self.backend_swagger_client.get_model('GitInfoDTO')
        GitCommitDTO = self.backend_swagger_client.get_model('GitCommitDTO')

        git_info_data = None
        if git_info is not None:
            git_info_data = GitInfoDTO(
                commit=GitCommitDTO(
                    commitId=git_info.commit_id,
                    message=git_info.message,
                    authorName=git_info.author_name,
                    authorEmail=git_info.author_email,
                    commitDate=git_info.commit_date
                ),
                remotes=git_info.remote_urls,
                currentBranch=git_info.active_branch,
                repositoryDirty=git_info.repository_dirty
            )

        try:
            params = ExperimentCreationParams(
                projectId=project.internal_id,
                name=name,
                description=description,
                parameters=self._convert_to_api_parameters(params),
                properties=self._convert_to_api_properties(properties),
                tags=tags,
                gitInfo=git_info_data,
                enqueueCommand="command",  # legacy (it's ignored but any non-empty string is required)
                entrypoint=entrypoint,
                execArgsTemplate="",  # legacy,
                abortable=abortable,
                monitored=monitored,
                hostname=hostname,
                notebookId=notebook_id,
                checkpointId=checkpoint_id
            )

            kwargs = {
                'experimentCreationParams': params,
                'X-Neptune-CliVersion': self.client_lib_version
            }
            api_experiment = self.backend_swagger_client.api.createExperiment(**kwargs).response().result

            return self._convert_to_experiment(api_experiment, project)
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)
        except HTTPBadRequest as e:
            error_type = extract_response_field(e.response, 'type')
            if error_type == 'DUPLICATE_PARAMETER':
                raise ExperimentValidationError('Parameter list contains duplicates.')
            elif error_type == 'INVALID_TAG':
                raise ExperimentValidationError(extract_response_field(e.response, 'message'))
            else:
                raise
        except HTTPUnprocessableEntity as e:
            if extract_response_field(e.response, 'type') == 'LIMIT_OF_EXPERIMENTS_IN_PROJECT_REACHED':
                raise ExperimentLimitReached()
            else:
                raise

    def upload_source_code(self, experiment, source_target_pairs):
        upload_source_entries = [
            UploadEntry(source_path, target_path)
            for source_path, target_path in source_target_pairs
        ]
        upload_to_storage(upload_entries=upload_source_entries,
                          upload_api_fun=self.upload_experiment_source,
                          upload_tar_api_fun=self.extract_experiment_source,
                          warn_limit=100 * 1024 * 1024,
                          experiment=experiment)

    @with_api_exceptions_handler
    def get_notebook(self, project, notebook_id):
        try:
            api_notebook_list = self.leaderboard_swagger_client.api.listNotebooks(
                projectIdentifier=project.internal_id,
                id=[notebook_id]
            ).response().result

            if not api_notebook_list.entries:
                raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

            api_notebook = api_notebook_list.entries[0]

            return Notebook(
                backend=self,
                project=project,
                _id=api_notebook.id,
                owner=api_notebook.owner
            )
        except HTTPNotFound:
            raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

    @with_api_exceptions_handler
    def get_last_checkpoint(self, project, notebook_id):
        try:
            api_checkpoint_list = self.leaderboard_swagger_client.api.listCheckpoints(
                notebookId=notebook_id,
                offset=0,
                limit=1
            ).response().result

            if not api_checkpoint_list.entries:
                raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

            checkpoint = api_checkpoint_list.entries[0]
            return Checkpoint(checkpoint.id, checkpoint.name, checkpoint.path)
        except HTTPNotFound:
            raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

    @with_api_exceptions_handler
    def create_notebook(self, project):
        try:
            api_notebook = self.leaderboard_swagger_client.api.createNotebook(
                projectIdentifier=project.internal_id
            ).response().result

            return Notebook(
                backend=self,
                project=project,
                _id=api_notebook.id,
                owner=api_notebook.owner
            )
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

    @with_api_exceptions_handler
    def create_checkpoint(self, notebook_id, jupyter_path, _file=None):
        if _file is not None:
            with self._upload_raw_data(api_method=self.leaderboard_swagger_client.api.createCheckpoint,
                                       data=_file,
                                       headers={"Content-Type": "application/octet-stream"},
                                       path_params={
                                           "notebookId": notebook_id
                                       },
                                       query_params={
                                           "jupyterPath": jupyter_path
                                       }) as response:
                if response.status_code == NOT_FOUND:
                    raise NotebookNotFound(notebook_id=notebook_id)
                else:
                    response.raise_for_status()
                    CheckpointDTO = self.leaderboard_swagger_client.get_model('CheckpointDTO')
                    return CheckpointDTO.unmarshal(response.json())
        else:
            try:
                NewCheckpointDTO = self.leaderboard_swagger_client.get_model('NewCheckpointDTO')
                return self.leaderboard_swagger_client.api.createEmptyCheckpoint(
                    notebookId=notebook_id,
                    checkpoint=NewCheckpointDTO(path=jupyter_path)
                ).response().result
            except HTTPNotFound:
                return None

    @with_api_exceptions_handler
    def get_experiment(self, experiment_id):
        return self.backend_swagger_client.api.getExperiment(experimentId=experiment_id).response().result

    @with_api_exceptions_handler
    def update_experiment(self, experiment, properties):
        EditExperimentParams = self.backend_swagger_client.get_model('EditExperimentParams')
        KeyValueProperty = self.backend_swagger_client.get_model('KeyValueProperty')
        try:
            self.backend_swagger_client.api.updateExperiment(
                experimentId=experiment.internal_id,
                editExperimentParams=EditExperimentParams(
                    properties=[KeyValueProperty(
                        key=key,
                        value=properties[key]
                    ) for key in properties]
                )
            ).response()
            return experiment
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id
            )

    @with_api_exceptions_handler
    def set_property(self, experiment, key, value):
        properties = {p.key: p.value for p in self.get_experiment(experiment.internal_id).properties}
        properties[key] = str(value)
        return self.update_experiment(
            experiment=experiment,
            properties=properties
        )

    @with_api_exceptions_handler
    def remove_property(self, experiment, key):
        properties = {p.key: p.value for p in self.get_experiment(experiment.internal_id).properties}
        del properties[key]
        return self.update_experiment(
            experiment=experiment,
            properties=properties
        )

    @with_api_exceptions_handler
    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        UpdateTagsParams = self.backend_swagger_client.get_model('UpdateTagsParams')
        try:
            self.backend_swagger_client.api.updateTags(
                updateTagsParams=UpdateTagsParams(
                    experimentIds=[experiment.internal_id],
                    groupsIds=[],
                    tagsToAdd=tags_to_add,
                    tagsToDelete=tags_to_delete
                )
            ).response().result
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id
            )
        except HTTPBadRequest as e:
            error_type = extract_response_field(e.response, 'type')
            if error_type == 'INVALID_TAG':
                raise ExperimentValidationError(extract_response_field(e.response, 'message'))
            else:
                raise

    @handle_quota_limits
    def upload_experiment_source(self, experiment, data, progress_indicator):
        self._upload_loop(partial(self._upload_raw_data,
                                  api_method=self.backend_swagger_client.api.uploadExperimentSource),
                          data=data,
                          progress_indicator=progress_indicator,
                          path_params={'experimentId': experiment.internal_id},
                          query_params={})

    @handle_quota_limits
    def extract_experiment_source(self, experiment, data):
        return self._upload_tar_data(
            experiment=experiment,
            api_method=self.backend_swagger_client.api.uploadExperimentSourceAsTarstream,
            data=data
        )

    @with_api_exceptions_handler
    def create_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        ChannelParams = self.backend_swagger_client.get_model('ChannelParams')

        try:
            params = ChannelParams(
                name=name,
                channelType=channel_type
            )

            channel = self.backend_swagger_client.api.createChannel(
                experimentId=experiment.internal_id,
                channelToCreate=params
            ).response().result

            return self._convert_channel_to_channel_with_last_value(channel)
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)
        except HTTPConflict:
            raise ChannelAlreadyExists(channel_name=name, experiment_short_id=experiment.id)

    @with_api_exceptions_handler
    def get_channels(self, experiment) -> Dict[str, object]:
        api_experiment = self.get_experiment(experiment.internal_id)
        channels_last_values_by_name = dict((ch.channelName, ch) for ch in api_experiment.channelsLastValues)
        channels = dict()
        for ch in api_experiment.channels:
            last_value = channels_last_values_by_name.get(ch.name, None)
            if last_value is not None:
                ch.x = last_value.x
                ch.y = last_value.y
            elif ch.lastX is not None:
                ch.x = ch.lastX
                ch.y = None
            else:
                ch.x = None
                ch.y = None
            channels[ch.name] = ch
        return channels

    @with_api_exceptions_handler
    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        try:
            self.backend_swagger_client.api.resetChannel(
                id=channel_id
            ).response()
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ChannelNotFound(channel_id)

    @with_api_exceptions_handler
    def create_system_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        ChannelParams = self.backend_swagger_client.get_model('ChannelParams')

        try:
            params = ChannelParams(
                name=name,
                channelType=channel_type
            )

            channel = self.backend_swagger_client.api.createSystemChannel(
                experimentId=experiment.internal_id,
                channelToCreate=params
            ).response().result

            return self._convert_channel_to_channel_with_last_value(channel)
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)
        except HTTPConflict:
            raise ChannelAlreadyExists(channel_name=name, experiment_short_id=experiment.id)

    @with_api_exceptions_handler
    def get_system_channels(self, experiment) -> Dict[str, object]:
        try:
            channels = self.backend_swagger_client.api.getSystemChannels(
                experimentId=experiment.internal_id,
            ).response().result

            return {ch.name: ch for ch in channels}
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def send_channels_values(self, experiment, channels_with_values):
        InputChannelValues = self.backend_swagger_client.get_model('InputChannelValues')
        Point = self.backend_swagger_client.get_model('Point')
        Y = self.backend_swagger_client.get_model('Y')

        input_channels_values = []
        for channel_with_values in channels_with_values:
            points = [Point(
                timestampMillis=int(value.ts * 1000.0),
                x=value.x,
                y=Y(numericValue=value.y.get('numeric_value'),
                    textValue=value.y.get('text_value'),
                    inputImageValue=value.y.get('image_value'))
            ) for value in channel_with_values.channel_values]

            input_channels_values.append(InputChannelValues(
                channelId=channel_with_values.channel_id,
                values=points
            ))

        try:
            batch_errors = self.backend_swagger_client.api.postChannelValues(
                experimentId=experiment.internal_id,
                channelsValues=input_channels_values
            ).response().result

            if batch_errors:
                raise ChannelsValuesSendBatchError(experiment.id, batch_errors)
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def mark_succeeded(self, experiment):
        CompletedExperimentParams = self.backend_swagger_client.get_model('CompletedExperimentParams')

        try:
            self.backend_swagger_client.api.markExperimentCompleted(
                experimentId=experiment.internal_id,
                completedExperimentParams=CompletedExperimentParams(
                    state='succeeded',
                    traceback=''  # FIXME
                )
            ).response()

            return experiment
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)
        except HTTPUnprocessableEntity:
            raise ExperimentAlreadyFinished(experiment.id)

    @with_api_exceptions_handler
    def mark_failed(self, experiment, traceback):
        CompletedExperimentParams = self.backend_swagger_client.get_model('CompletedExperimentParams')

        try:
            self.backend_swagger_client.api.markExperimentCompleted(
                experimentId=experiment.internal_id,
                completedExperimentParams=CompletedExperimentParams(
                    state='failed',
                    traceback=traceback
                )
            ).response()

            return experiment
        except HTTPNotFound as e:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id) from e
        except HTTPUnprocessableEntity:
            raise ExperimentAlreadyFinished(experiment.id)

    @with_api_exceptions_handler
    def ping_experiment(self, experiment):
        try:
            self.backend_swagger_client.api.pingExperiment(experimentId=experiment.internal_id).response()
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def create_hardware_metric(self, experiment, metric):
        SystemMetricParams = self.backend_swagger_client.get_model('SystemMetricParams')

        try:
            series = [gauge.name() for gauge in metric.gauges]
            system_metric_params = SystemMetricParams(
                name=metric.name, description=metric.description,
                resourceType=metric.resource_type.replace("MEMORY", "RAM"),
                unit=metric.unit, min=metric.min_value, max=metric.max_value, series=series)

            metric_dto = self.backend_swagger_client.api.createSystemMetric(
                experimentId=experiment.internal_id, metricToCreate=system_metric_params
            ).response().result

            return metric_dto.id
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        SystemMetricValues = self.backend_swagger_client.get_model('SystemMetricValues')
        SystemMetricPoint = self.backend_swagger_client.get_model('SystemMetricPoint')

        try:
            metrics_by_name = {metric.name: metric for metric in metrics}

            system_metric_values = [
                SystemMetricValues(
                    metricId=metrics_by_name.get(report.metric.name).internal_id,
                    seriesName=gauge_name,
                    values=[
                        SystemMetricPoint(
                            timestampMillis=int(metric_value.timestamp * 1000.0),
                            x=int(metric_value.running_time * 1000.0),
                            y=metric_value.value
                        )
                        for metric_value in metric_values
                    ]
                )
                for report in metric_reports
                for gauge_name, metric_values in groupby(report.values, lambda value: value.gauge_name)
            ]

            response = self.backend_swagger_client.api.postSystemMetricValues(
                experimentId=experiment.internal_id, metricValues=system_metric_values).response()

            return response
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    def log_artifact(self, experiment, artifact, destination=None):
        if isinstance(artifact, str):
            if os.path.exists(artifact):
                target_name = os.path.basename(artifact) if destination is None else destination
                upload_entry = UploadEntry(os.path.abspath(artifact), normalize_file_name(target_name))
            else:
                raise FileNotFound(artifact)
        elif hasattr(artifact, 'read'):
            if destination is not None:
                upload_entry = UploadEntry(artifact, normalize_file_name(destination))
            else:
                raise ValueError("destination is required for file streams")
        else:
            raise ValueError("Artifact must be a local path or an IO object")

        upload_to_storage(upload_entries=[upload_entry],
                          upload_api_fun=self._upload_experiment_output,
                          upload_tar_api_fun=self._extract_experiment_output,
                          experiment=experiment)

    def delete_artifacts(self, experiment, path):
        if path is None:
            raise ValueError("path argument must not be None")

        paths = path
        if not isinstance(path, list):
            paths = [path]
        for path in paths:
            if path is None:
                raise ValueError("path argument must not be None")
            normalized_path = os.path.normpath(path)
            if normalized_path.startswith(".."):
                raise ValueError("path to delete must be within project's directory")
            if normalized_path == "." or normalized_path == "/" or not normalized_path:
                raise ValueError("Cannot delete whole artifacts directory")
        try:
            for path in paths:
                self.rm_data(experiment=experiment, path=path)
        except PathInProjectNotFound:
            raise FileNotFound(path)

    @with_api_exceptions_handler
    def rm_data(self, experiment, path):
        try:
            return self.backend_swagger_client.api.deleteExperimentOutput(
                experimentIdentifier=str(experiment.internal_id),
                path=path).response().result
        except BaseException as e:
            print("exception was raised: ", e)
            raise e

    @with_api_exceptions_handler
    def download_data(self, experiment, path, destination):
        project_storage_path = "/{exp_id}/output/{file}".format(exp_id=experiment.id, file=path)
        project = experiment._project  # pylint: disable=protected-access
        with self._download_raw_data(api_method=self.backend_swagger_client.api.downloadData,
                                     headers={"Accept": "application/octet-stream"},
                                     path_params={},
                                     query_params={
                                         "projectId": project.internal_id,
                                         "path": project_storage_path
                                     }) as response:
            if response.status_code == NOT_FOUND:
                raise PathInProjectNotFound(path=path, project_identifier=project.full_id)
            else:
                response.raise_for_status()

            with open(destination, "wb") as f:
                for chunk in response.iter_content(chunk_size=10 * 1024 * 1024):
                    if chunk:
                        f.write(chunk)

    def download_sources(self, experiment, path=None, destination_dir=None):
        if not path:
            path = ""
        destination_dir = assure_directory_exists(destination_dir)

        download_request = self.prepare_source_download_request(experiment, path)
        self.download_from_request(download_request, destination_dir, path)

    @with_api_exceptions_handler
    def prepare_source_download_request(self, experiment, path):
        try:
            return self.backend_swagger_client.api.prepareForDownload(
                experimentIdentity=experiment.internal_id,
                resource='source',
                path=path
            ).response().result
        except HTTPNotFound:
            raise FileNotFound(path)

    def download_artifacts(self, experiment: Experiment, path=None, destination_dir=None):
        if not path:
            path = ""
        destination_dir = assure_directory_exists(destination_dir)

        download_request = self.prepare_output_download_request(experiment, path)
        self.download_from_request(download_request, destination_dir, path)

    def download_artifact(self, experiment: Experiment, path=None, destination_dir=None):
        destination_dir = assure_directory_exists(destination_dir)
        destination_path = os.path.join(destination_dir, os.path.basename(path))

        try:
            self.download_data(experiment, path, destination_path)
        except PathInProjectNotFound as e:
            raise FileNotFound(path) from e

    @with_api_exceptions_handler
    def prepare_output_download_request(self, experiment, path):
        try:
            return self.backend_swagger_client.api.prepareForDownload(
                experimentIdentity=experiment.internal_id,
                resource='output',
                path=path
            ).response().result
        except HTTPNotFound:
            raise FileNotFound(path)

    def download_from_request(self, download_request, destination_dir, path):
        sleep_time = 1
        max_sleep_time = 16
        while not hasattr(download_request, "downloadUrl"):
            time.sleep(sleep_time)
            sleep_time = min(sleep_time * 2, max_sleep_time)
            download_request = self.get_download_request(download_request.id)

        ssl_verify = True
        if os.getenv("NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"):
            ssl_verify = False

        # We do not use ApiClient here cause `downloadUrl` can be any url (not only Neptune API endpoint)
        response = requests.get(
            url=download_request.downloadUrl,
            headers={"Accept": "application/zip"},
            stream=True,
            verify=ssl_verify
        )

        with response:
            filename = None
            if 'content-disposition' in response.headers:
                content_disposition = response.headers['content-disposition']
                filenames = re.findall("filename=(.+)", content_disposition)
                if filenames:
                    filename = filenames[0]

            if not filename:
                filename = os.path.basename(path.rstrip("/")) + ".zip"

            destination_path = os.path.join(destination_dir, filename)
            with open(destination_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=10 * 1024 * 1024):
                    if chunk:
                        f.write(chunk)

    @with_api_exceptions_handler
    def get_download_request(self, request_id):
        return self.backend_swagger_client.api.getDownloadRequest(id=request_id).response().result

    @handle_quota_limits
    def _upload_experiment_output(self, experiment, data, progress_indicator):
        self._upload_loop(partial(self._upload_raw_data,
                                  api_method=self.backend_swagger_client.api.uploadExperimentOutput),
                          data=data,
                          progress_indicator=progress_indicator,
                          path_params={'experimentId': experiment.internal_id},
                          query_params={})

    @handle_quota_limits
    def _extract_experiment_output(self, experiment, data):
        return self._upload_tar_data(
            experiment=experiment,
            api_method=self.backend_swagger_client.api.uploadExperimentOutputAsTarstream,
            data=data
        )

    @staticmethod
    def _get_all_items(get_portion, step):
        items = []

        previous_items = None
        while previous_items is None or len(previous_items) >= step:
            previous_items = get_portion(limit=step, offset=len(items))
            items += previous_items

        return items

    def _get_parameter_with_type(self, parameter):
        string_type = 'string'
        double_type = 'double'
        if isinstance(parameter, bool):
            return (string_type, str(parameter))
        elif isinstance(parameter, float) or isinstance(parameter, int):
            if math.isinf(parameter) or math.isnan(parameter):
                return (string_type, json.dumps(parameter))
            else:
                return (double_type, str(parameter))
        else:
            return (string_type, str(parameter))

    def _convert_to_api_parameters(self, raw_params):
        Parameter = self.backend_swagger_client.get_model('Parameter')

        params = []
        for name, value in raw_params.items():
            (parameter_type, string_value) = self._get_parameter_with_type(value)
            params.append(
                Parameter(
                    id=str(uuid.uuid4()),
                    name=name,
                    parameterType=parameter_type,
                    value=string_value
                )
            )

        return params

    def _convert_to_api_properties(self, raw_properties):
        KeyValueProperty = self.backend_swagger_client.get_model('KeyValueProperty')

        return [
            KeyValueProperty(
                key=key,
                value=str(value)
            ) for key, value in raw_properties.items()
        ]

    def _convert_to_experiment(self, api_experiment, project):
        # pylint: disable=protected-access
        return Experiment(backend=project._backend,
                          project=project,
                          _id=api_experiment.shortId,
                          internal_id=api_experiment.id)

    def _convert_channel_to_channel_with_last_value(self, channel):
        ChannelWithValueDTO = self.leaderboard_swagger_client.get_model('ChannelWithValueDTO')
        return ChannelWithLastValue(
            ChannelWithValueDTO(
                channelId=channel.id,
                channelName=channel.name,
                channelType=channel.channelType,
                x=None,
                y=None
            )
        )

    def _upload_loop(self, fun, data, progress_indicator, **kwargs):
        ret = None
        for part in data.generate():
            ret = with_api_exceptions_handler(self._upload_loop_chunk)(fun, part, data, **kwargs)
            progress_indicator.progress(part.end - part.start)

        data.close()
        return ret

    def _upload_loop_chunk(self, fun, part, data, **kwargs):
        if data.length is not None:
            binary_range = "bytes=%d-%d/%d" % (part.start, part.end - 1, data.length)
        else:
            binary_range = "bytes=%d-%d" % (part.start, part.end - 1)
        headers = {
            "Content-Type": "application/octet-stream",
            "Content-Filename": data.filename,
            "X-Range": binary_range,
        }
        if data.permissions is not None:
            headers["X-File-Permissions"] = data.permissions
        response = fun(data=part.get_data(), headers=headers, **kwargs)
        response.raise_for_status()
        return response

    def _upload_raw_data(self, api_method, data, headers, path_params, query_params):
        url = self.api_address + api_method.operation.path_name + "?"

        for key, val in path_params.items():
            url = url.replace("{" + key + "}", val)

        for key, val in query_params.items():
            url = url + key + "=" + val + "&"

        session = self.http_client.session

        request = self.authenticator.apply(
            requests.Request(
                method='POST',
                url=url,
                data=data,
                headers=headers
            )
        )

        return session.send(session.prepare_request(request))

    def _download_raw_data(self, api_method, headers, path_params, query_params):
        url = self.api_address + api_method.operation.path_name + "?"

        for key, val in path_params.items():
            url = url.replace("{" + key + "}", val)

        for key, val in query_params.items():
            url = url + key + "=" + val + "&"

        session = self.http_client.session

        request = self.authenticator.apply(
            requests.Request(
                method='GET',
                url=url,
                headers=headers
            )
        )

        return session.send(session.prepare_request(request), stream=True)

    @with_api_exceptions_handler
    def _upload_tar_data(self, experiment, api_method, data):
        url = self.api_address + api_method.operation.path_name
        url = url.replace("{experimentId}", experiment.internal_id)

        session = self.http_client.session

        request = self.authenticator.apply(
            requests.Request(
                method='POST',
                url=url,
                data=io.BytesIO(data),
                headers={
                    "Content-Type": "application/octet-stream"
                }
            )
        )

        response = session.send(session.prepare_request(request))
        response.raise_for_status()
        return response



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import re
from typing import List, Optional, Dict, Iterable, Tuple, Any

import click
from bravado.exception import HTTPNotFound, HTTPUnprocessableEntity
from simplejson import JSONDecodeError

from neptune.new.internal.artifacts.types import ArtifactFileData
from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.new.exceptions import (
    ClientHttpError,
    FetchAttributeNotFoundException,
    RunNotFound,
    RunUUIDNotFound,
    InternalClientError,
    MetadataInconsistency,
    NeptuneException,
    NeptuneLegacyProjectException,
    ProjectNotFound,
    ProjectNameCollision,
    NeptuneStorageLimitException,
    ArtifactNotFoundException,
    NeptuneFeaturesNotAvailableException,
)
from neptune.new.internal.backends.api_model import (
    ApiRun,
    ArtifactAttribute,
    Attribute,
    AttributeType,
    AttributeWithProperties,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatPointValue,
    FloatSeriesAttribute,
    FloatSeriesValues,
    ImageSeriesValues,
    IntAttribute,
    LeaderboardEntry,
    Project,
    StringAttribute,
    StringPointValue,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
    Workspace,
)
from neptune.new.internal.backends.hosted_file_operations import (
    download_file_attribute,
    download_file_set_attribute,
    download_image_series_element,
    upload_file_attribute,
    upload_file_set_attribute,
)
from neptune.new.internal.backends.hosted_artifact_operations import (
    track_to_new_artifact,
    track_to_existing_artifact
)
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.backends.operation_api_name_visitor import OperationApiNameVisitor
from neptune.new.internal.backends.operation_api_object_converter import OperationApiObjectConverter
from neptune.new.internal.backends.operations_preprocessor import OperationsPreprocessor
from neptune.new.internal.backends.utils import (
    with_api_exceptions_handler,
    MissingApiClient,
    OptionalFeatures,
)
from neptune.new.internal.backends.hosted_client import (
    DEFAULT_REQUEST_KWARGS,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import (
    Operation,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.new.internal.utils import verify_type, base64_decode
from neptune.new.internal.utils.generic_attribute_mapper import map_attribute_result_to_value
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.internal.backends.utils import build_operation_url, ssl_verify
from neptune.new.internal.websockets.websockets_factory import WebsocketsFactory
from neptune.new.types.atoms import GitRef
from neptune.new.version import version as neptune_client_version

_logger = logging.getLogger(__name__)


class HostedNeptuneBackend(NeptuneBackend):
    def __init__(self, credentials: Credentials, proxies: Optional[Dict[str, str]] = None):
        self.credentials = credentials
        self.proxies = proxies
        self.missing_features = []

        self._http_client, self._client_config = create_http_client_with_auth(
            credentials=credentials,
            ssl_verify=ssl_verify(),
            proxies=proxies
        )

        self.backend_client = create_backend_client(self._client_config, self._http_client)
        self.leaderboard_client = create_leaderboard_client(self._client_config, self._http_client)

        try:
            self.artifacts_client = create_artifacts_client(self._client_config, self._http_client)
        except JSONDecodeError:
            # thanks for nice error handling, bravado
            self.artifacts_client = MissingApiClient(self)
            self.missing_features.append(OptionalFeatures.ARTIFACTS)

    def verify_feature_available(self, feature_name: str):
        if feature_name in self.missing_features:
            raise NeptuneFeaturesNotAvailableException(self.missing_features)

    def get_display_address(self) -> str:
        return self._client_config.display_url

    def websockets_factory(self, project_id: str, run_id: str) -> Optional[WebsocketsFactory]:
        base_url = re.sub(r'^http', 'ws', self._client_config.api_url)
        return WebsocketsFactory(
            url=build_operation_url(
                base_url,
                f'/api/notifications/v1/runs/{project_id}/{run_id}/signal'
            ),
            session=self._http_client.authenticator.auth.session,
            proxies=self.proxies
        )

    def close(self) -> None:
        pass

    @with_api_exceptions_handler
    def get_project(self, project_id: str) -> Project:
        verify_type("project_id", project_id, str)

        project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, project_id)
        workspace, name = project_spec['workspace'], project_spec['project']

        try:
            if not workspace:
                available_projects = list(filter(lambda p: p.name == name,
                                                 self.get_available_projects(search_term=name)))

                if len(available_projects) == 1:
                    project = available_projects[0]
                    project_id = f'{project.workspace}/{project.name}'
                elif len(available_projects) > 1:
                    raise ProjectNameCollision(
                        project_id=project_id,
                        available_projects=available_projects
                    )
                else:
                    raise ProjectNotFound(
                        project_id=project_id,
                        available_projects=self.get_available_projects(),
                        available_workspaces=self.get_available_workspaces()
                    )

            response = self.backend_client.api.getProject(
                projectIdentifier=project_id,
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            warning = response.metadata.headers.get('X-Server-Warning')
            if warning:
                click.echo(warning)  # TODO print in color once colored exceptions are added
            project = response.result
            project_version = project.version if hasattr(project, 'version') else 1
            if project_version < 2:
                raise NeptuneLegacyProjectException(project_id)
            return Project(project.id, project.name, project.organizationName)
        except HTTPNotFound:
            raise ProjectNotFound(project_id,
                                  available_projects=self.get_available_projects(workspace_id=workspace),
                                  available_workspaces=list() if workspace else self.get_available_workspaces())

    @with_api_exceptions_handler
    def get_available_projects(self,
                               workspace_id: Optional[str] = None,
                               search_term: Optional[str] = None
                               ) -> List[Project]:
        try:
            response = self.backend_client.api.listProjects(
                limit=5,
                organizationIdentifier=workspace_id,
                searchTerm=search_term,
                sortBy=['lastViewed'],
                sortDirection=['descending'],
                userRelation='memberOrHigher',
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            warning = response.metadata.headers.get('X-Server-Warning')
            if warning:
                click.echo(warning)  # TODO print in color once colored exceptions are added
            projects = response.result.entries
            return list(map(
                lambda project: Project(project.id, project.name, project.organizationName),
                projects))
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_available_workspaces(self) -> List[Workspace]:
        try:
            response = self.backend_client.api.listOrganizations(
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            warning = response.metadata.headers.get('X-Server-Warning')
            if warning:
                click.echo(warning)  # TODO print in color once colored exceptions are added
            workspaces = response.result
            return list(map(
                lambda workspace: Workspace(_id=workspace.id, name=workspace.name),
                workspaces))
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_run(self, run_id: str):
        try:
            run = self.leaderboard_client.api.getExperiment(
                experimentId=run_id,
                **DEFAULT_REQUEST_KWARGS,
            ).response().result
            return ApiRun(run.id, run.shortId, run.organizationName, run.projectName, run.trashed)
        except HTTPNotFound:
            raise RunNotFound(run_id)

    @with_api_exceptions_handler
    def create_run(self,
                   project_id: str,
                   git_ref: Optional[GitRef] = None,
                   custom_run_id: Optional[str] = None,
                   notebook_id: Optional[str] = None,
                   checkpoint_id: Optional[str] = None
                   ) -> ApiRun:
        verify_type("project_id", project_id, str)

        git_info = {
            "commit": {
                "commitId": git_ref.commit_id,
                "message": git_ref.message,
                "authorName": git_ref.author_name,
                "authorEmail": git_ref.author_email,
                "commitDate": git_ref.commit_date
            },
            "repositoryDirty": git_ref.dirty,
            "currentBranch": git_ref.branch,
            "remotes": git_ref.remotes
        } if git_ref else None

        params = {
            "projectIdentifier": project_id,
            "cliVersion": str(neptune_client_version),
            "gitInfo": git_info,
            "customId": custom_run_id,
        }

        if notebook_id is not None and checkpoint_id is not None:
            params["notebookId"] = notebook_id if notebook_id is not None else None
            params["checkpointId"] = checkpoint_id if checkpoint_id is not None else None

        kwargs = {
            'experimentCreationParams': params,
            'X-Neptune-CliVersion': str(neptune_client_version),
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            run = self.leaderboard_client.api.createExperiment(**kwargs).response().result
            return ApiRun(run.id, run.shortId, run.organizationName, run.projectName, run.trashed)
        except HTTPNotFound:
            raise ProjectNotFound(project_id=project_id)

    @with_api_exceptions_handler
    def create_checkpoint(self, notebook_id: str, jupyter_path: str) -> Optional[str]:
        try:
            return self.leaderboard_client.api.createEmptyCheckpoint(
                notebookId=notebook_id,
                checkpoint={
                    "path": jupyter_path
                },
                **DEFAULT_REQUEST_KWARGS,
            ).response().result.id
        except HTTPNotFound:
            return None

    @with_api_exceptions_handler
    def ping_run(self, run_id: str):
        request_kwargs = {
            "_request_options": {
                "timeout": 10, "connect_timeout": 10,
            }
        }
        try:
            self.leaderboard_client.api.ping(
                experimentId=run_id,
                **request_kwargs,
            ).response().result
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id)

    def execute_operations(self, run_id: str, operations: List[Operation]) -> List[NeptuneException]:
        errors = []

        operations_preprocessor = OperationsPreprocessor()
        operations_preprocessor.process(operations)
        errors.extend(operations_preprocessor.get_errors())

        upload_operations, artifact_operations, other_operations = [], [], []

        for op in operations_preprocessor.get_operations():
            if isinstance(
                    op,
                    (UploadFile, UploadFileContent, UploadFileSet)
            ):
                upload_operations.append(op)
            elif isinstance(op, TrackFilesToArtifact):
                artifact_operations.append(op)
            else:
                other_operations.append(op)

        if artifact_operations:
            self.verify_feature_available(OptionalFeatures.ARTIFACTS)

        # Upload operations should be done first since they are idempotent
        errors.extend(
            self._execute_upload_operations_with_400_retry(
                run_id=run_id,
                upload_operations=upload_operations)
        )

        artifact_operations_errors, assign_artifact_operations = self._execute_artifact_operations(
            run_id=run_id,
            artifact_operations=artifact_operations
        )

        errors.extend(artifact_operations_errors)
        other_operations.extend(assign_artifact_operations)

        if other_operations:
            errors.extend(self._execute_operations(run_id, other_operations))

        return errors

    def _execute_upload_operations(self,
                                   run_id: str,
                                   upload_operations: List[Operation]) -> List[NeptuneException]:
        errors = list()

        for op in upload_operations:
            if isinstance(op, UploadFile):
                error = upload_file_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    source=op.file_path,
                    ext=op.ext)
                if error:
                    errors.append(error)
            elif isinstance(op, UploadFileContent):
                error = upload_file_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    source=base64_decode(op.file_content),
                    ext=op.ext)
                if error:
                    errors.append(error)
            elif isinstance(op, UploadFileSet):
                error = upload_file_set_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    file_globs=op.file_globs,
                    reset=op.reset)
                if error:
                    errors.append(error)
            else:
                raise InternalClientError("Upload operation in neither File or FileSet")

        return errors

    def _execute_upload_operations_with_400_retry(
            self,
            run_id: str,
            upload_operations: List[Operation]) -> List[NeptuneException]:
        while True:
            try:
                return self._execute_upload_operations(run_id, upload_operations)
            except ClientHttpError as ex:
                if "Length of stream does not match given range" not in ex.response:
                    raise ex

    @with_api_exceptions_handler
    def _execute_artifact_operations(
            self,
            run_id: str,
            artifact_operations: List[TrackFilesToArtifact]
    ) -> Tuple[List[Optional[NeptuneException]], List[Optional[Operation]]]:
        errors = list()
        assign_operations = list()

        for op in artifact_operations:
            try:
                artifact_hash = self.get_artifact_attribute(run_id, op.path).hash
            except FetchAttributeNotFoundException:
                artifact_hash = None

            try:
                if artifact_hash is None:
                    assign_operation = track_to_new_artifact(
                        swagger_client=self.artifacts_client,
                        project_id=op.project_id,
                        path=op.path,
                        parent_identifier=run_id,
                        entries=op.entries,
                        default_request_params=DEFAULT_REQUEST_KWARGS
                    )
                else:
                    assign_operation = track_to_existing_artifact(
                        swagger_client=self.artifacts_client,
                        project_id=op.project_id,
                        path=op.path,
                        artifact_hash=artifact_hash,
                        parent_identifier=run_id,
                        entries=op.entries,
                        default_request_params=DEFAULT_REQUEST_KWARGS
                    )

                if assign_operation:
                    assign_operations.append(assign_operation)
            except NeptuneException as error:
                errors.append(error)

        return errors, assign_operations

    @with_api_exceptions_handler
    def _execute_operations(self,
                            run_id: str,
                            operations: List[Operation]) -> List[MetadataInconsistency]:
        kwargs = {
            'experimentId': run_id,
            'operations': [{
                'path': path_to_str(op.path),
                OperationApiNameVisitor().visit(op): OperationApiObjectConverter().convert(op)
            } for op in operations],
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            result = self.leaderboard_client.api.executeOperations(**kwargs).response().result
            return [MetadataInconsistency(err.errorDescription) for err in result]
        except HTTPNotFound as e:
            raise RunUUIDNotFound(run_id=run_id) from e
        except HTTPUnprocessableEntity:
            raise NeptuneStorageLimitException()

    @with_api_exceptions_handler
    def get_attributes(self, run_id: str) -> List[Attribute]:
        def to_attribute(attr) -> Attribute:
            return Attribute(attr.name, AttributeType(attr.type))

        params = {
            'experimentId': run_id,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            run = self.leaderboard_client.api.getExperimentAttributes(**params).response().result

            attribute_type_names = [at.value for at in AttributeType]
            accepted_attributes = [attr for attr in run.attributes if attr.type in attribute_type_names]

            # Notify about ignored attrs
            ignored_attributes = \
                set(attr.type for attr in run.attributes) - set(attr.type for attr in accepted_attributes)
            if ignored_attributes:
                _logger.warning(
                    "Ignored following attributes (unknown type): %s.\n"
                    "Try to upgrade `neptune-client.", ignored_attributes
                )

            return [to_attribute(attr) for attr in accepted_attributes if attr.type in attribute_type_names]
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id=run_id)

    def download_file_series_by_index(self, run_id: str, path: List[str],
                                      index: int, destination: str):
        try:
            download_image_series_element(
                swagger_client=self.leaderboard_client,
                run_id=run_id,
                attribute=path_to_str(path),
                index=index,
                destination=destination
            )
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    def download_file(self, run_id: str, path: List[str], destination: Optional[str] = None):
        try:
            download_file_attribute(
                swagger_client=self.leaderboard_client,
                run_id=run_id,
                attribute=path_to_str(path),
                destination=destination)
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    def download_file_set(self, run_id: str, path: List[str], destination: Optional[str] = None):
        download_request = self._get_file_set_download_request(run_id, path)
        try:
            download_file_set_attribute(
                swagger_client=self.leaderboard_client,
                download_id=download_request.id,
                destination=destination)
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    @with_api_exceptions_handler
    def get_float_attribute(self, run_id: str, path: List[str]) -> FloatAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFloatAttribute(**params).response().result
            return FloatAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_int_attribute(self, run_id: str, path: List[str]) -> IntAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getIntAttribute(**params).response().result
            return IntAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_bool_attribute(self, run_id: str, path: List[str]) -> BoolAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getBoolAttribute(**params).response().result
            return BoolAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_file_attribute(self, run_id: str, path: List[str]) -> FileAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFileAttribute(**params).response().result
            return FileAttribute(name=result.name, ext=result.ext, size=result.size)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_attribute(self, run_id: str, path: List[str]) -> StringAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringAttribute(**params).response().result
            return StringAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_datetime_attribute(self, run_id: str, path: List[str]) -> DatetimeAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getDatetimeAttribute(**params).response().result
            return DatetimeAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_artifact_attribute(self, run_id: str, path: List[str]) -> ArtifactAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getArtifactAttribute(**params).response().result
            return ArtifactAttribute(
                hash=result.hash
            )
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def list_artifact_files(self, project_id: str, artifact_hash: str) -> List[ArtifactFileData]:
        params = {
            'projectIdentifier': project_id,
            'hash': artifact_hash,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.artifacts_client.api.listArtifactFiles(**params).response().result
            return [
                ArtifactFileData.from_dto(a) for a in result.files
            ]
        except HTTPNotFound:
            raise ArtifactNotFoundException(artifact_hash)

    @with_api_exceptions_handler
    def get_float_series_attribute(self, run_id: str, path: List[str]) -> FloatSeriesAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFloatSeriesAttribute(**params).response().result
            return FloatSeriesAttribute(result.last)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_attribute(self, run_id: str, path: List[str]) -> StringSeriesAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSeriesAttribute(**params).response().result
            return StringSeriesAttribute(result.last)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_set_attribute(self, run_id: str, path: List[str]) -> StringSetAttribute:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSetAttribute(**params).response().result
            return StringSetAttribute(set(result.values))
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_image_series_values(self, run_id: str, path: List[str],
                                offset: int, limit: int) -> ImageSeriesValues:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            'limit': limit,
            'offset': offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getImageSeriesValues(**params).response().result
            return ImageSeriesValues(result.totalItemCount)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_values(self, run_id: str, path: List[str],
                                 offset: int, limit: int) -> StringSeriesValues:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            'limit': limit,
            'offset': offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSeriesValues(**params).response().result
            return StringSeriesValues(result.totalItemCount,
                                      [StringPointValue(v.timestampMillis, v.step, v.value) for v in result.values])
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_float_series_values(self, run_id: str, path: List[str],
                                offset: int, limit: int) -> FloatSeriesValues:
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            'limit': limit,
            'offset': offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFloatSeriesValues(**params).response().result
            return FloatSeriesValues(result.totalItemCount,
                                     [FloatPointValue(v.timestampMillis, v.step, v.value) for v in result.values])
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def fetch_atom_attribute_values(self, run_id: str, path: List[str]) -> List[Tuple[str, AttributeType, Any]]:
        params = {
            'experimentId': run_id,
        }
        try:
            namespace_prefix = path_to_str(path)
            if namespace_prefix:
                # don't want to catch "ns/attribute/other" while looking for "ns/attr"
                namespace_prefix += "/"
            result = self.leaderboard_client.api.getExperimentAttributes(**params).response().result
            return [
                (attr.name, attr.type, map_attribute_result_to_value(attr))
                for attr in result.attributes if attr.name.startswith(namespace_prefix)
            ]
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id)

    @with_api_exceptions_handler
    def _get_file_set_download_request(self, run_id: str, path: List[str]):
        params = {
            'experimentId': run_id,
            'attribute': path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            return self.leaderboard_client.api.prepareForDownloadFileSetAttributeZip(**params).response().result
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_leaderboard(self, project_id: str,
                        _id: Optional[Iterable[str]] = None,
                        state: Optional[Iterable[str]] = None,
                        owner: Optional[Iterable[str]] = None,
                        tags: Optional[Iterable[str]] = None
                        ) -> List[LeaderboardEntry]:

        def get_portion(limit, offset):
            return self.leaderboard_client.api.getLeaderboard(
                projectIdentifier=project_id,
                shortId=_id, state=state, owner=owner, tags=tags, tagsMode='and',
                sortBy=['shortId'], sortFieldType=['string'], sortDirection=['ascending'],
                limit=limit, offset=offset,
                **DEFAULT_REQUEST_KWARGS,
            ).response().result.entries

        def to_leaderboard_entry(entry) -> LeaderboardEntry:
            supported_attribute_types = {item.value for item in AttributeType}
            attributes: List[AttributeWithProperties] = []
            for attr in entry.attributes:
                if attr.type in supported_attribute_types:
                    properties = attr.__getitem__("{}Properties".format(attr.type))
                    attributes.append(AttributeWithProperties(
                        attr.name,
                        AttributeType(attr.type),
                        properties
                    ))
            return LeaderboardEntry(entry.id, attributes)

        try:
            return [to_leaderboard_entry(e) for e in self._get_all_items(get_portion, step=100)]
        except HTTPNotFound:
            raise ProjectNotFound(project_id)

    def get_run_url(self, run_id: str, workspace: str, project_name: str, short_id: str) -> str:
        base_url = self.get_display_address()
        return f"{base_url}/{workspace}/{project_name}/e/{short_id}"

    @staticmethod
    def _get_all_items(get_portion, step):
        items = []
        previous_items = None
        while previous_items is None or len(previous_items) >= step:
            previous_items = get_portion(limit=step, offset=len(items))
            items += previous_items
        return items



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pathlib
import typing
from datetime import datetime
from urllib.parse import urlparse

from neptune.new.exceptions import (
    NeptuneLocalStorageAccessException,
    NeptuneUnsupportedArtifactFunctionalityException,
)
from neptune.new.internal.artifacts.file_hasher import FileHasher
from neptune.new.internal.artifacts.types import ArtifactDriver, ArtifactFileData, ArtifactFileType


class LocalArtifactDriver(ArtifactDriver):
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    @staticmethod
    def get_type() -> str:
        return ArtifactFileType.LOCAL.value

    @classmethod
    def matches(cls, path: str) -> bool:
        return urlparse(path).scheme in ('file', '')

    @classmethod
    def _serialize_metadata(cls, metadata: typing.Dict[str, typing.Any]) -> typing.Dict[str, str]:
        return {
            "file_path": metadata['file_path'],
            "last_modified": datetime.fromtimestamp(metadata['last_modified']).strftime(cls.DATETIME_FORMAT),
        }

    @classmethod
    def _deserialize_metadata(cls, metadata: typing.Dict[str, str]) -> typing.Dict[str, typing.Any]:
        return {
            "file_path": metadata['file_path'],
            "last_modified": datetime.strptime(metadata['last_modified'], cls.DATETIME_FORMAT),
        }

    @classmethod
    def get_tracked_files(cls, path: str, destination: str = None) -> typing.List[ArtifactFileData]:
        file_protocol_prefix = 'file://'
        if path.startswith(file_protocol_prefix):
            path = path[len(file_protocol_prefix):]

        if '*' in path:
            raise NeptuneUnsupportedArtifactFunctionalityException(
                f'Wildcard characters (*,?) in location URI ({path}) are not supported.'
            )

        source_location = pathlib.Path(path).expanduser()

        stored_files: typing.List[ArtifactFileData] = list()

        files_to_check = source_location.rglob('*') if source_location.is_dir() else [source_location]
        for file in files_to_check:
            # symlink dirs are omitted by rglob('*')
            if not file.is_file():
                continue

            if source_location.is_dir():
                file_path = file.relative_to(source_location).as_posix()
            else:
                file_path = file.name
            file_path = file_path if destination is None else (pathlib.Path(destination) / file_path).as_posix()

            stored_files.append(
                ArtifactFileData(
                    file_path=file_path,
                    file_hash=FileHasher.get_local_file_hash(file),
                    type=ArtifactFileType.LOCAL.value,
                    size=file.stat().st_size,
                    metadata=cls._serialize_metadata({
                        'file_path': f'file://{file.resolve().as_posix()}',
                        'last_modified': file.stat().st_mtime,
                    })
                )
            )

        return stored_files

    @classmethod
    def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
        parsed_path = urlparse(file_definition.metadata.get("file_path"))
        absolute_path = pathlib.Path(parsed_path.netloc + parsed_path.path)

        if not absolute_path.is_file():
            raise NeptuneLocalStorageAccessException(
                path=absolute_path,
                expected_description="an existing file"
            )

        os.makedirs(str(destination.parent), exist_ok=True)
        if destination.exists():
            os.remove(destination)
        destination.symlink_to(absolute_path)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
import time
from io import BytesIO
from typing import List, Optional, Dict, Iterable, Callable, Set, Union
from urllib.parse import urlencode

from bravado.client import SwaggerClient
from bravado.exception import HTTPUnprocessableEntity
from bravado.requests_client import RequestsClient
from requests import Request, Response

from neptune.new.exceptions import FileUploadError, MetadataInconsistency, InternalClientError, \
    NeptuneException, NeptuneStorageLimitException
from neptune.new.internal.backends.utils import build_operation_url, with_api_exceptions_handler
from neptune.new.internal.utils import get_absolute_paths, get_common_root
from neptune.internal.storage.datastream import compress_to_tar_gz_in_memory, FileChunkStream, FileChunk
from neptune.internal.storage.storage_utils import scan_unique_upload_entries, split_upload_files, UploadEntry, \
    normalize_file_name


def upload_file_attribute(swagger_client: SwaggerClient,
                          run_id: str,
                          attribute: str,
                          source: Union[str, bytes],
                          ext: str
                          ) -> Optional[NeptuneException]:
    if isinstance(source, str) and not os.path.isfile(source):
        return FileUploadError(source, "Path not found or is a not a file.")

    target = attribute
    if ext:
        target += "." + ext

    try:
        url = build_operation_url(
            swagger_client.swagger_spec.api_url,
            swagger_client.api.uploadAttribute.operation.path_name
        )
        upload_entry = UploadEntry(source if isinstance(source, str) else BytesIO(source), target)
        _upload_loop(file_chunk_stream=FileChunkStream(upload_entry),
                     response_handler=_attribute_upload_response_handler,
                     http_client=swagger_client.swagger_spec.http_client,
                     url=url,
                     query_params={
                         "experimentId": run_id,
                         "attribute": attribute,
                         "ext": ext
                     })
    except MetadataInconsistency as e:
        return e


def upload_file_set_attribute(swagger_client: SwaggerClient,
                              run_id: str,
                              attribute: str,
                              file_globs: Iterable[str],
                              reset: bool,
                              ) -> Optional[NeptuneException]:
    unique_upload_entries = get_unique_upload_entries(file_globs)

    try:
        for package in split_upload_files(unique_upload_entries):
            if package.is_empty() and not reset:
                continue

            uploading_multiple_entries = package.len > 1
            creating_a_single_empty_dir = package.len == 1 and not package.items[0].is_stream() \
                                          and os.path.isdir(package.items[0].source_path)

            if uploading_multiple_entries or creating_a_single_empty_dir or package.is_empty():
                data = compress_to_tar_gz_in_memory(upload_entries=package.items)
                url = build_operation_url(
                    swagger_client.swagger_spec.api_url,
                    swagger_client.api.uploadFileSetAttributeTar.operation.path_name
                )
                result = upload_raw_data(http_client=swagger_client.swagger_spec.http_client,
                                         url=url,
                                         data=BytesIO(data),
                                         headers={"Content-Type": "application/octet-stream"},
                                         query_params={
                                             "experimentId": run_id,
                                             "attribute": attribute,
                                             "reset": str(reset)
                                         })
                _attribute_upload_response_handler(result)
            else:
                url = build_operation_url(
                    swagger_client.swagger_spec.api_url,
                    swagger_client.api.uploadFileSetAttributeChunk.operation.path_name
                )
                file_chunk_stream = FileChunkStream(package.items[0])
                _upload_loop(file_chunk_stream=file_chunk_stream,
                             response_handler=_attribute_upload_response_handler,
                             http_client=swagger_client.swagger_spec.http_client,
                             url=url,
                             query_params={
                                 "experimentId": str(run_id),
                                 "attribute": attribute,
                                 "reset": str(reset),
                                 "path": file_chunk_stream.filename
                             })

            reset = False
    except MetadataInconsistency as e:
        return e


def get_unique_upload_entries(file_globs: Iterable[str]) -> Set[UploadEntry]:
    absolute_paths = get_absolute_paths(file_globs)
    common_root = get_common_root(absolute_paths)

    upload_entries: List[UploadEntry] = []
    if common_root is not None:
        for absolute_path in absolute_paths:
            upload_entries.append(UploadEntry(absolute_path, normalize_file_name(
                os.path.relpath(absolute_path, common_root))))
    else:
        for absolute_path in absolute_paths:
            upload_entries.append(UploadEntry(absolute_path, normalize_file_name(absolute_path)))

    return scan_unique_upload_entries(upload_entries)


def _attribute_upload_response_handler(result: bytes) -> None:
    try:
        parsed = json.loads(result)
    except json.JSONDecodeError:
        raise InternalClientError("Unexpected response from server: {}".format(result))

    if isinstance(parsed, type(None)):
        return
    if isinstance(parsed, dict):
        if "errorDescription" in parsed:
            raise MetadataInconsistency(parsed["errorDescription"])
        else:
            raise InternalClientError("Unexpected response from server: {}".format(result))


def _upload_loop(file_chunk_stream: FileChunkStream, response_handler: Callable[[bytes], None], **kwargs):
    for chunk in file_chunk_stream.generate():
        result = _upload_loop_chunk(chunk, file_chunk_stream, **kwargs)
        response_handler(result)

    file_chunk_stream.close()


def _upload_loop_chunk(chunk: FileChunk, file_chunk_stream: FileChunkStream, **kwargs):
    if file_chunk_stream.length is not None:
        binary_range = "bytes=%d-%d/%d" % (chunk.start, chunk.end - 1, file_chunk_stream.length)
    else:
        binary_range = "bytes=%d-%d" % (chunk.start, chunk.end - 1)
    headers = {
        "Content-Filename": file_chunk_stream.filename,
        "X-Range": binary_range,
        "Content-Type": "application/octet-stream"
    }
    if file_chunk_stream.permissions is not None:
        headers["X-File-Permissions"] = file_chunk_stream.permissions
    return upload_raw_data(data=chunk.get_data(), headers=headers, **kwargs)


@with_api_exceptions_handler
def upload_raw_data(http_client: RequestsClient,
                    url: str,
                    data,
                    path_params: Optional[Dict[str, str]] = None,
                    query_params: Optional[Dict[str, str]] = None,
                    headers: Optional[Dict[str, str]] = None):
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(Request(method='POST', url=url, data=data, headers=headers))

    response = session.send(session.prepare_request(request))
    if response.status_code == HTTPUnprocessableEntity.status_code:
        raise NeptuneStorageLimitException()
    response.raise_for_status()
    return response.content


def download_image_series_element(swagger_client: SwaggerClient,
                                  run_id: str,
                                  attribute: str,
                                  index: int,
                                  destination: str):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.getImageSeriesValue.operation.path_name
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={},
        query_params={"experimentId": run_id, "attribute": attribute, "index": index})
    _store_response_as_file(response, os.path.join(destination, "{}.{}"
                                                   .format(index, response.headers['content-type'].split('/')[-1])))


def download_file_attribute(swagger_client: SwaggerClient,
                            run_id: str,
                            attribute: str,
                            destination: Optional[str] = None):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.downloadAttribute.operation.path_name
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={"Accept": "application/octet-stream"},
        query_params={"experimentId": run_id, "attribute": attribute})
    _store_response_as_file(response, destination)


def download_file_set_attribute(swagger_client: SwaggerClient,
                                download_id: str,
                                destination: Optional[str] = None):
    download_url: Optional[str] = _get_download_url(swagger_client, download_id)
    next_sleep = 0.5
    while download_url is None:
        time.sleep(next_sleep)
        next_sleep = min(2 * next_sleep, 5)
        download_url = _get_download_url(swagger_client, download_id)

    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=download_url,
        headers={"Accept": "application/zip"})
    _store_response_as_file(response, destination)


def _get_download_url(swagger_client: SwaggerClient, download_id: str):
    params = {"id": download_id}
    download_request = swagger_client.api.getDownloadPrepareRequest(**params).response().result
    return download_request.downloadUrl


def _store_response_as_file(response: Response, destination: Optional[str] = None):
    if destination is None:
        target_file = _get_content_disposition_filename(response)
    elif os.path.isdir(destination):
        target_file = os.path.join(destination, _get_content_disposition_filename(response))
    else:
        target_file = destination
    with response:
        with open(target_file, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)


def _get_content_disposition_filename(response: Response) -> str:
    content_disposition = response.headers['Content-Disposition']
    return content_disposition[content_disposition.rfind("filename=") + 9:].strip('"')


@with_api_exceptions_handler
def _download_raw_data(http_client: RequestsClient,
                       url: str,
                       path_params: Optional[Dict[str, str]] = None,
                       query_params: Optional[Dict[str, str]] = None,
                       headers: Optional[Dict[str, str]] = None
                       ) -> Response:
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(Request(method='GET', url=url, headers=headers))

    response = session.send(session.prepare_request(request), stream=True)
    response.raise_for_status()
    return response


def _generate_url(url: str,
                  path_params: Optional[Dict[str, str]] = None,
                  query_params: Optional[Dict[str, str]] = None,
                  ) -> str:
    for key, val in (path_params or dict()).items():
        url = url.replace("{" + key + "}", val)
    if query_params:
        url = url + "?" + urlencode(list(query_params.items()))
    return url



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import threading

from neptune.internal.notebooks.comm import send_checkpoint_created
from neptune.utils import is_ipython

_logger = logging.getLogger(__name__)

_checkpoints_lock = threading.Lock()
_checkpoints = dict()


def create_checkpoint(backend, notebook_id, notebook_path):
    if is_ipython():
        # pylint:disable=bad-option-value,import-outside-toplevel,import-error
        import IPython
        ipython = IPython.core.getipython.get_ipython()
        execution_count = -1
        if ipython.kernel is not None:
            execution_count = ipython.kernel.execution_count
        with _checkpoints_lock:

            if execution_count in _checkpoints:
                return _checkpoints[execution_count]

            checkpoint = backend.create_checkpoint(notebook_id, notebook_path)
            if ipython is not None and ipython.kernel is not None:
                send_checkpoint_created(notebook_id=notebook_id,
                                        notebook_path=notebook_path,
                                        checkpoint_id=checkpoint.id)
                _checkpoints[execution_count] = checkpoint
            return checkpoint
